/*
 * This is acrate3d.c only included at the top of scat.c
 */

double trapz(double *xx, double *yy, int n1, int n2) {
  int i;
  double s=0.;
  for(i = n1; i <= n2-1; i++) {
	s += .5*(yy[i+1]+yy[i])*(xx[i+1]-xx[i]);
  }
  return s;
}

double get_wq(double q, int lt) {   /* phonon frequency */
  if (q > QMAX) {
	if (lt == LA) lt = LO;   // LA outside BZ extends into LO
	q = 2.*QMAX - q;         // and flip it back to first BZ
  } else if (q < 0.) {
	q = -q;
  }
  // here lt = 1 (LA), 2 (TA), 3 (LO), 4 (TO) as in constants.h
  return (LTAO[lt-1][0]*q*q + LTAO[lt-1][1]*q + LTAO[lt-1][2]);
}

double get_nq(double q, int lt) {   /* phonon occupation */
  double xx = hbar*get_wq(q,lt)/(kb*T);
  /* if (xx < 3.5) return (1./xx -0.5 +xx/12. -xx*xx*xx/720.);
	 else return exp(-xx); */
  return (1./(exp(xx)-1.));
}

double get_Mq(double q, int lt, double aed) {
  double gq;
  if (q > QMAX) q = 2.*QMAX-q;    // returns to the first BZ
#ifdef POP
  gq = 3.*(sin(q*Rws)-q*Rws*cos(q*Rws))/pow(q*Rws,3.);
#else
  gq = 1.0;   // overlap integral
#endif
  return (q*q*gq*gq*(get_nq(q,lt)+0.5+0.5*aed)/get_wq(q,lt));
}

double fabs_cost(double q, double k, double E, int lt, double aed) {
  /*** |cos(t)| = |+/-q/(2k) + m*w(q)/(hbar*k*q)|   +em/-ab  ***/
  double wq = get_wq(q,lt);
  // aed = -1 for ABS and +1 for EMS
  return fabs(aed*q/(2.*k) +
			  mdos*(1. + alpha*(2.*E - aed*hbar*wq))*wq/(hbar*k*q));
}

double find_root(double q1,double q2,double k,double E,int lt,double aed) {
  /* find root of |cos(t)| - 1 in interval q1..q2 */
  double f1, f2, fs, q=q2, eps = (q2-q1)/40.;  // 20.
  f1 = fabs_cost(q1,k,E,lt,aed);
  f2 = fabs_cost(q2,k,E,lt,aed);
  fs = (f2-f1)/fabs(f2-f1);
  while (q2-q1 > eps) {
	q = .5*(q1+q2);
	if (fs*(fabs_cost(q,k,E,lt,aed)- 1.) > 0.) q2 = q;
	else q1 = q;
  }
  if (fs < 0.) return q2;
  else return q1;
}

double acrate3(double E, int lt, double aed) {
  /****** aed = -1. for absorption, aed = +1. for emission ******/
  int i, i1, i2, nloc, N=200, MULT=20;   // 100 and 10
  double *q, *integ, q1, q2, dq, k, acost, C3, G, DP;
  q=vector(1,MULT*N);  integ=vector(1,MULT*N);

  /* k is the electron wavevector in Herring-Vogt space */
  k = sqrt(2.*mdos*E*(1.+alpha*E))/hbar;

  q1 = QMIN;  q2 = QMAX;
  i1 = 0;  i2 = N;
  dq = (QMAX-QMIN)/((double) (N-1));
  q[1] = QMIN;
  
  for (i = 1; i <= N; i++) {
	if (i >= 2) q[i] = q[i-1]+dq;
	/***  cos(t) = +/- q/(2k) + m*w(q)/(hbar*k*q)   +em/-ab  ***/
	acost = fabs_cost(q[i],k,E,lt,aed);
	/* find range of q's [q1..q2] where the cosine is non-zero */
	if ((acost <= 1.) && (i1 == 0)) {
	  i1 = i;
	  if (i >= 2) q1 = find_root(q[i-1],q[i],k,E,lt,aed);
	} else if ((acost >  1.) && (i2 == N) && (i1)) {
	  i2 = i-1;
	  q2 = find_root(q[i-1],q[i],k,E,lt,aed);
	}
  }

  if (i1 == 0) {             /* in case no q-values were found */
	G = 0.;
  } else {                   /* found q-range to integrate over */
	if (i2 == N) {
	  printf("\n*** Hit BZ End while integrating acoustic rates ");
	  printf("at E=%g.  Reduce EMAX! ***\n",E);
	  // or continue at your own risk (continuing works, sort of)
	  exit(0);
	}
	//nloc = MULT*((int) floor(q2-q1)/dq);
	nloc = MULT*ifloor((q2-q1)/dq);
	dq = (q2-q1)/((double) (nloc-1));
	
	q[1] = q1;
	/* compute q-dependent part of the integrand */
	for (i = 1; i <= nloc; i++) {
	  if (i >= 2) q[i] = q[i-1]+dq;
	  integ[i] = q[i]*get_Mq(q[i],lt,aed);
	}
	
	if (lt == LA) {  /* longitudinal */
	  DP = DLA;
	} else {         /* transverse */
	  DP = DTA;
	}

	C3 = DP*DP*mdos/(4.*pi*rho*hbar*hbar*k);
	G = C3*trapz(q,integ,1,nloc);
  }
  
  free_vector(q,1,MULT*N);
  free_vector(integ,1,MULT*N);

  return G;
}
