/*
  Auxilliary functions I will need along the way
*/

#include <stdlib.h>
#include <stdio.h>
#include <mathimf.h>
//#include <math.h>
#include "constants.h"
#include "bsd-strlcat.c"
#include "bsd-strlcpy.c"
#include <assert.h>

__inline int ifloor(double x) {
  return ((int) floor(x));
}

__inline int iround(double x) {
  return ((int) round(x));
}

__inline void nrerror(char error_text[]) {
  fprintf(stderr,"%s\n",error_text);
  exit(-1);
}

double *vector(int nl, int nh) {
/* allocate a double vector with subscript range v[nl..nh] */
  int i;  double *v;
  v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  //for (i = 0; i <= (nh-nl); i++) v[i] = 0.0;
  if (!v) nrerror("allocation failure in vector()");
  return v-nl+NR_END;
}

int *ivector(int nl, int nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
  int *v;
  v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
  //for (i = 0; i <= (n2-n1); i++) v[i] = 0;
  if (!v) nrerror("allocation failure in ivector()");
  return v-nl+NR_END;
}

/*int *ivector(int n1, int n2) {
  int *v, i;
  v = (int *) malloc((n2-n1+1)*sizeof(int));
  for (i = 0; i <= (n2-n1); i++) v[i] = 0;
  if (!v) {  printf("\nError allocating memory for ivector.\n"); exit(-1);  }
  return (v-n1);
  }*/

double **matrix(int nr, int nc) {
  int r, nrow=nr+1, ncol=nc+1;
  double **m = malloc(nrow * sizeof(double *));
  assert(m != NULL);
  for (r = 0; r < nrow; r++) {
	m[r] = calloc( ncol, sizeof(double) );
	assert(m[r] != NULL);
  }
  return m;
}

/*double **matrix(int nrow, int ncol) {
  int row;
  double *tmp;
  double **m;  

  // Allocate Vector of Row Pointers
  m = (double **) malloc( nrow * sizeof(double *) );
  if (m == NULL) {
	fprintf(stderr, "Error. FILE: %s LINE: %d\n", __FILE__, __LINE__);
	return(NULL);  
  }

  // Allocate Contiguous Block of Memory for the Matrix
  // Use 'calloc()' so that the memory is returned cleared.
  tmp = (double *) calloc( nrow*ncol, sizeof(double) );
  if (tmp == NULL) {
	free(m);
	fprintf(stderr, "Error. FILE: %s LINE: %d\n", __FILE__, __LINE__);
	return(NULL);
  }

  // Loop through the Block & Assign Pointers
  for (row = 0; row < nrow; row++) {
	m[row] = &tmp[row * ncol];
  }

  return(m);
  }*/

/*double **matrix(int m1, int m2, int n1, int n2) {
  int i, nrow=m2-m1+1, ncol=n2-n1+1;
  double **m;
  
  m=(double **) malloc((size_t)((nrow+NR_END)*sizeof(double*)));
  if (!m) nrerror("allocation failure 1 in matrix()");
  m += NR_END;
  m -= m1;
  
  m[m1]=(double *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(double)));
  if (!m[m1]) nrerror("allocation failure 2 in matrix()");
  m[m1] += NR_END;
  m[m1] -= n1;
  
  for(i=m1+1;i<=m2;i++) m[i]=m[i-1]+ncol;
  return m;
}*/

void printvector(double *v, int n1, int n2, char *dir, char *fname) {
  int i; FILE *fp;
  char dirfile[MAXWORD];
  snprintf(dirfile, MAXWORD, "%s/%s", dir, fname);
  fp = fopen(dirfile, "w");
  for (i = n1; i <= n2; i++) {
	fprintf(fp, "%g\n", v[i]);
  }
  fclose(fp);
}

void printivector(int *v, int n1, int n2, char *dir, char *fname) {
  int i; FILE *fp;
  char dirfile[MAXWORD];
  snprintf(dirfile, MAXWORD, "%s/%s", dir, fname);
  fp = fopen(dirfile, "w");
  for (i = n1; i <= n2; i++) {
	fprintf(fp, "%i\n", v[i]);
  }
  fclose(fp);
}

void print2vectors(double *v, double *w, int n1, int n2, char *dir, char *fname) {
  int i; FILE *fp;
  char dirfile[MAXWORD];
  snprintf(dirfile, MAXWORD, "%s/%s", dir, fname);
  fp = fopen(dirfile, "w");
  for (i = n1; i <= n2; i++) {
	fprintf(fp, "%g\t%g\n", v[i], w[i]);
  }
  fclose(fp);
}

void printmatrix(double **m, int m1, int m2, int n1, int n2, char *dir,
				 char *fname, int be) {
  int i, j; FILE *fp;
  char dirfile[MAXWORD];
  snprintf(dirfile, MAXWORD, "%s/%s", dir, fname);
  fp = fopen(dirfile, "w");
  if (be == 2) {
	fprintf(fp,"%%   X [cm]       Ne [1/cm^3]       E [eV]        Vx [cm/s]        Pot [V]        Fld [V/cm]\n");
  }
  for (i = m1; i <= m2; i++) {
	if (be == 1) fprintf(fp, "%g\t", ebins[i]);
	else if (be == 2) fprintf(fp, "% .6e ", Xgrid[i]);
	for (j = n1; j <= n2; j++) {
	  fprintf(fp, "% .8e ", m[i][j]);
	}
	fprintf(fp, "\n");
  }
  fclose(fp);
}

void printgridavgs(double time) {
  int i;  FILE *fp;
  char dirfile[MAXWORD];

  snprintf(dirfile, MAXWORD, "%s/%s", OUTDIR, "avgs.txt");
  fp = fopen(dirfile, "w");

  fprintf(fp,"%% Time elapsed **beyond** TTRANS = %g\n", time);
  fprintf(fp,"%%   X [cm]     N [1/cm^3]     E [eV]      Vx [cm/s]    Pot [V]     Fld [V/cm]    AcPhon       OpPhon\n");

  for (i = 1; i <= nxg; i++) {
	fprintf(fp, "% .5e % .5e % .5e % .5e % .5e % .5e % .5e % .5e\n",
			Xgrid[i], realsimscale*GridAvg[i][NEg]/DAgrid[i], GridAvg[i][EEg],
			GridAvg[i][VXg], GridAvg[i][POg], GridAvg[i][FFg],
			realsimscale*Agrid[i]/time/DAgrid[i], realsimscale*Ogrid[i]/time/DAgrid[i]);
  }
  fclose(fp);
}

double meanmatrixcol(double **m, int n1, int n2, int ed) {
  int i; double mean=0.0;
  for (i = n1; i <= n2; i++) mean += m[ed][i];
  return (mean/((double) (n2-n1+1)));
}

double sumvector(double *v, int n1, int n2) {
  int i; double sum=0.0;
  for (i = n1; i <= n2; i++) sum += v[i];
  return sum;
}

double meanvector(double *v, int n1, int n2) {
  int i; double mean=0.0;
  for (i = n1; i <= n2; i++) mean += v[i];
  return (mean/((double) (n2-n1+1)));
}

double normvector(double *v, int n1, int n2) {
  int i; double norm=0.0;
  for (i = n1; i <= n2; i++) norm += v[i]*v[i];
  return (sqrt(norm));
}

double stdvector(double *v, int n1, int n2) {
  int i;
  double std=0.0, mean = meanvector(v,n1,n2);
  for (i = n1; i <= n2; i++) std += (v[i]-mean)*(v[i]-mean);
  return (sqrt(std)/((double) (n2-n1+1)));
}

void zerovector(double *v, int n1, int n2) {
  int i;
  for (i = n1; i <= n2; i++) v[i] = 0.0;
}

void free_vector(double *v, int nl, int nh) {
/* free a double vector allocated with vector() */
  free((FREE_ARG) (v+nl-NR_END));
}

/* calculate gamma given momentum */
__inline double get_gamma(int j) {
  assert ((j >= 1) && (j <= nelec));
  return (.5*((Elec[PX][j]-pvx[iv[j]])*(Elec[PX][j]-pvx[iv[j]])/mx[iv[j]] +
			  (Elec[PY][j]-pvy[iv[j]])*(Elec[PY][j]-pvy[iv[j]])/my[iv[j]] +
			  (Elec[PZ][j]-pvz[iv[j]])*(Elec[PZ][j]-pvz[iv[j]])/mz[iv[j]]));
}

/* calculate energy given momentum */
__inline double energy(int j) {
  /* tomizawa page 8, jacoboni (rev.mod) page 668 */
  assert ((j >= 1) && (j <= nelec));
  return (o2alpha*(sqrt(1.+ alpha4*get_gamma(j)) - 1.));
}

__inline double velocity(double mm, double pp, double ee) {
  /* tomizawa page 10, jacoboni (rev.mod) page 669 */
  return (pp/(mm*(1.+ alpha2*ee)));
}

double ran1(void) {
  printf("don't use me!  RAND_MAX = 32767 under AIX !?\n");
  printf("the random() manpage claims it returns ints between\n0 and 2e31-1 but that doesn't seem to be the case at all\n");
  exit(-1);
  return ((double) random())/((double) RAND_MAX);
}

int locate(double *v, int n1, int n2, double num) {
  /* bisection routine -- given vector v w/index range n1..n2, find which
	 index j (n1<=j<=n2) is the LEFT end of interval v[j] <= num < v[j+1];
     it returns n2 if num >= v[n2] */
  int ju,jm,jl,j;

  jl=n1-1; ju=n2+1;           /* set lower and upper limits */
  while (ju-jl > 1) {
	jm=(ju+jl) >> 1;          /* compute midpoint */
	if (num >= v[jm]) jl=jm;  /* replace either lower or upper limit */
	else ju=jm;
  }
  if (num == v[n1]) j=n1;
  else if (num == v[n2]) j=n2-1;
  else j=jl;

  return j;
}

int locate_nearest(double *v, int n, double num) {
  /* returns which index range 1-n the number num is NEAREST to */
  int j = locate(v,1,n-1,num);  /* 1<=j<=n-1 as v[j]<=num<=v[j+1] */
  if ((num < v[j]) || (num > v[j+1])) {
	printf("damn!\n");
	exit(-1);
  }
  if ((num-v[j]) < (v[j+1]-num)) return j;
  else return (j+1);
}

void cloud_in_cell(int i, int ixcic, int iycic) {
  double xcic, ycic, oAAcic;
  xcic = Elec[X][i]-Xgrid[ixcic];
  if (DIM == 1) {
	wcic[2] = xcic/DXgrid[ixcic];

	// to replace cloud-in-cell with nearest-neighbor scheme
	//if (wcic[2] > 0.5) wcic[2] = 1.0;
	//else wcic[2] = 0.0;

	wcic[1] = 1. - wcic[2];
	return;
  }
  /* return weights (all < 1) for the 4 nearest-neighbors in wcic[]
	 recall the numbering scheme for the 4 corners of the cell:
        (ix-1)*nyg+iy  1 *---------* 3  ix*nyg+iy
                         |         |
                         |         |
      (ix-1)*nyg+iy+1  2 *---------* 4  ix*nyg+iy+1
  */
  else {
	ycic = Elec[Y][i]-Ygrid[iycic];
	// oAAgrid is the inverse cell area (stored ahead in initial.c)
	oAAcic = oAAgrid[(ixcic-1)*nyg+iycic];
	wcic[1] = (DXgrid[ixcic]-xcic)*(DYgrid[iycic]-ycic)*oAAcic;
	wcic[2] = (DXgrid[ixcic]-xcic)*ycic*oAAcic;
	wcic[3] = xcic*(DYgrid[iycic]-ycic)*oAAcic;
	wcic[4] = xcic*ycic*oAAcic;
  }
}

void update_avg(double *VA, int ix, int iy, double x) {
  int iua;
  if (DIM == 1) {
	VA[ix] += wcic[1]*x;
	VA[ix+1] += wcic[2]*x;
  } else {
	iua = (ix-1)*nyg+iy;
	VA[iua] += wcic[1]*x;
	VA[iua+1] += wcic[2]*x;
	VA[iua+nyg] += wcic[3]*x;
	VA[iua+nyg+1] += wcic[4]*x;
  }
}

void phonon_stats(double aed, double ephon, double *Ephacop, double *qGrd,
				  double *NxGrd, double *xGrd, int ipx, int ipy) {
  // aed = -1. for absorption and +1. for emission
  if (DIM) {
	update_avg(NxGrd, ipx, ipy, aed);
	update_avg(xGrd, ipx, ipy, aed*ephon);
  } else if (EFSTEPS == DIM) {     // store phonon event (0-D only)
	qGrd[1 + ifloor(ephon/DEph)] += aed;
	//qGrd[1 + ifloor(ephon/DEph)] += 1.;
  }
  (*Ephacop) += aed*ephon;
}

__inline void change_equiv_valley(int j) {
  int axis;
  assert ((j >= 1) && (j <= nelec));
  axis = (iv[j] + 1)/2;
  iv[j] = 4*axis - iv[j] - 1;
}

__inline void change_noneq_valley(int j) {
  int nv;
  double r = ranmc();
  assert ((j >= 1) && (j <= nelec));
  if ((iv[j]+1)/2 == 1) {          // i = 1 or 2
	nv = 3 + ifloor(4.*r);             // --> {3,4,5,6}
  } else if ((iv[j]+1)/2 == 2) {   // i = 3 or 4
	nv = 1 + ifloor(4.*r);
	if (nv > 2) nv += 2;               // --> {1,2}{5,6}
  } else {                         // i = 5 or 6
	nv = 1 + ifloor(4.*r);             // --> {1,2,3,4}
  }
  iv[j] = nv;
}

/*void check( void )
{
  int i,j;
  for ( i=1 ; i < NEBIN ; i++ )
	for ( j=1 ; j < NPROC ; j++ )
	  {
		char *elm_ptr = &(ScatMat[i][j]);
		char *mat_ptr = (char *)ScatMat;
		assert( elm_ptr  == mat_ptr + sizeof(double) * (j + NPROC*i) );
		assert( *elm_ptr == ScatMat[i][j] );
	  }
} */
