/*
 * This is boundary.c only included at the top of drift.c
 */
#include <math.h>
#include <stdlib.h>

int k_boundary_condition(int i) {
  double sign;
  int axis = (iv[i]+1)/2;               // (1,2);(3,4);(5,6) --> (1);(2);(3)
  int Pxyz = PX-1+axis;
  if (fabs(Elec[Pxyz][i]) > PQMAX) {
	// reflect back to the first Brillouin zone
	sign = iv[i]*2.-axis*4.+1.;         // (1,3,5) --> -1 and (2,4,6) --> +1
	Elec[Pxyz][i] -= sign*2.*PQMAX;
	// change valley index iv[] to opposite equivalent valley
	iv[i] = 4*axis - iv[i] - 1;
  }
  return 1;
}

//int boundary_condition(int i) {
int boundary_condition(int i, double oldX, double oldY, double vxold, double vyold) {
  //this is for when i get around to doing the time-reversal shtuff
  //bcgp = boundary_condition(j,vxold,vyold);
  double rnbc, xzR, theta1, temp;
  double XLCONT=OXLEF, XRCONT=OXRIG;
  int bcret = 1;
  int axis = (iv[i] + 1)/2;
  int Pxyz = PX-1+axis;

  // this is cludge-y but use the average electron temperature
  // just one grid node away from the contact so 2 or nxg-1

  /*************************************************************************
   ******* 1 DIM :: periodic boundaries w/thermal electron injection *******
   *************************************************************************/
  if (DIM == 1) {
	if (Elec[X][i] < XMIN) {
#ifdef RENEW
	  delete_particle(i);           // delete i.e. force exit
#else
	  //temp = max(GridAvg[30][EEg]/(1.5*kb), T);
	  create_electron(i,-2,T);  //TXgrid[nxg]);
	  // re-inject at the RIGHT contact
	  Elec[X][i] = XMAX + (Elec[X][i] - XMIN)*fabs(Elec[VX][i]/vxold);
#endif
	  if (STATS) CLEFT++;
	  bcret = 0;
	} else if (Elec[X][i] > XMAX) {
#ifdef RENEW
	  delete_particle(i);           // delete i.e. force exit
#else
	  //temp = max(GridAvg[30][EEg]/(1.5*kb), T);
	  //if (temp < 300) printf("temp = %g\n", temp);
	  create_electron(i,2,T);  //TXgrid[1]);
	  // re-inject at the LEFT contact
	  Elec[X][i] = XMIN + (Elec[X][i] - XMAX)*fabs(Elec[VX][i]/vxold);;
#endif
	  if (STATS) CRITE++;
	  bcret = 0;
	}
  } else {
	/***********************************************************************
	 ****** 2 DIM :: periodic boundaries w/thermal electron injection ******
	 ***********************************************************************/

	// REFLECT it specularly if it exists thru either side (left or right)
	if ((Elec[X][i] < XMIN) || (Elec[X][i] > XMAX)) {
	  if (Elec[X][i] < XMIN) {       /* exit LEFT ... */
		Elec[X][i] = 2.*XMIN - Elec[X][i];
	  } else {                       /* ... or RIGHT  */
		Elec[X][i] = 2.*XMAX - Elec[X][i];
	  }
	  Elec[VX][i] = -Elec[VX][i];
	  Elec[PX][i] = -Elec[PX][i];
	  if (axis == 1) change_equiv_valley(i);
	  bcret = 0;
	}

	// if it goes into the oxide, "above" the channel
	if ((Elec[Y][i] < CHTOP) && (Elec[X][i] > OXLEF) && (Elec[X][i] < OXRIG)) {
	  // case I, it came from the SOURCE, so reflect it back
	  if (oldX <= OXLEF) {
		Elec[X][i] = 2.*OXLEF - Elec[X][i];
		Elec[VX][i] = -Elec[VX][i];
		Elec[PX][i] = -Elec[PX][i];
		if (axis == 1) change_equiv_valley(i);
	  } else if (oldX >= OXRIG) {
		// case II, it came from the DRAIN, so reflect it back
		Elec[X][i] = 2.*OXRIG - Elec[X][i];
		Elec[VX][i] = -Elec[VX][i];
		Elec[PX][i] = -Elec[PX][i];
		if (axis == 1) change_equiv_valley(i);
	  } else if (oldY >= CHTOP) {
		// case III, it came from the CHANNEL, so scatter it back
		Elec[Y][i] = 2.*CHTOP - Elec[Y][i];
		Elec[VY][i] = -Elec[VY][i];
		Elec[PY][i] = -Elec[PY][i];
		if (axis == 2) change_equiv_valley(i);
		/******** time-reverse drift back to the interface *********/
		//this stuff isn't done yet
		/*tsover = (YMIN - Elec[Y][i])*2./(vyold + Elec[VY][i]);
		  Elec[X][i] -= .5*tsover*(vxold + Elec[VX][i]);
		  Elec[Y][i] = YMIN;
		  Elec[PX][i] += get_efield(i,1)*tsover;  // check the + sign
		  Elec[PY][i] += get_efield(i,2)*tsover;  // check the + sign
		  Elec[EE][i] = energy(j);
//		  Elec[VX][i] = velocity(mx[iv[i]],Elec[PX][i]-pvx[iv[i]],Elec[EE][i]);
//		  Elec[VY][i] = velocity(my[iv[i]],Elec[PY][i]-pvy[iv[i]],Elec[EE][i]);
		 */
		/************ crude model for surface scattering ***********/
		rnbc = ranmc();
		if (rnbc < SURFSCAT) {
		  do {
			xzR = sqrt((Elec[PX][i]-pvx[iv[i]])*(Elec[PX][i]-pvx[iv[i]])/mx[iv[i]]
					   + (Elec[PZ][i]-pvz[iv[i]])*(Elec[PZ][i]-pvz[iv[i]])/mz[iv[i]]);
			theta1 = 2.*pi*ranmc();
			Elec[PX][i] = pvx[iv[i]] + xzR*cos(theta1)*sqrt(mx[iv[i]]);
			Elec[PZ][i] = pvz[iv[i]] + xzR*sin(theta1)*sqrt(mz[iv[i]]);
			Elec[VX][i] = velocity(mx[iv[i]],Elec[PX][i]-pvx[iv[i]],Elec[EE][i]);
		  } while ((fabs(Elec[Pxyz][i]) > PQMAX) && (BZCLIP));
		}
		// drift2(i,tsover);
	  } else {
		printf("**** ERROR: oldX = %g    oldY = %g\n *****", oldX, oldY);
		exit(-1);
	  }
	  bcret = 0;
	}

	/******************************************************************
	 **************** If it exits through the CONTACTS ****************
	 ******************************************************************/
	if (Elec[Y][i] < YMIN) {
	  /********* adjust BOTH position and velocity+momentum! **********/
	  Elec[Y][i] = 2.*YMIN - Elec[Y][i];
	  if (Elec[X][i] < XLCONT) {         /* exit LEFT contact */
		Elec[X][i] = XRCONT + (XMAX-XRCONT)*ranmc();
		create_electron(i,8,T);      // re-inject w/positive flux
		if (STATS) CLEFT++;
	  } else if (Elec[X][i] > XRCONT) {  /* exit RIGHT contact */
		Elec[X][i] = XMIN + (XLCONT-XMIN)*ranmc();
		create_electron(i,8,T);      // re-inject w/positive flux
		if (STATS) CRITE++;
	  }
	  bcret = 0;
	}
	/******************************************************************
	 ************** REFLECT if it hits the BOTTOM oxide ***************
	 ******************************************************************/
	else if (Elec[Y][i] > CHBOT) {
	  /********* adjust both position and velocity+momentum! *****/
	  Elec[Y][i] = 2.*CHBOT - Elec[Y][i];
	  Elec[PY][i] = -Elec[PY][i];
	  Elec[VY][i] = -Elec[VY][i];
	  if (axis == 2) change_equiv_valley(i);
	  /************ crude model for surface scattering ***********/
	  rnbc = ranmc();
	  if (rnbc < SURFSCAT) {
		do {
		  xzR = sqrt((Elec[PX][i]-pvx[iv[i]])*(Elec[PX][i]-pvx[iv[i]])/mx[iv[i]]
					 + (Elec[PZ][i]-pvz[iv[i]])*(Elec[PZ][i]-pvz[iv[i]])/mz[iv[i]]);
		  theta1 = 2.*pi*ranmc();
		  Elec[PX][i] = pvx[iv[i]] + xzR*cos(theta1)*sqrt(mx[iv[i]]);
		  Elec[PZ][i] = pvz[iv[i]] + xzR*sin(theta1)*sqrt(mz[iv[i]]);
		  Elec[VX][i] = velocity(mx[iv[i]],Elec[PX][i]-pvx[iv[i]],Elec[EE][i]);
		} while ((fabs(Elec[Pxyz][i]) > PQMAX) && (BZCLIP));
	  }
	  bcret = 0;
	}
  }

#ifndef RENEW      // because in this case particles that exit are DELETED
  if (bcret == 0) {
	Gxp[i] = locate(Xgrid,1,nxg-1,Elec[X][i]);
	if (DIM == 2) Gyp[i] = locate(Ygrid,1,nyg-1,Elec[Y][i]);	
  }
#endif

  return bcret;
}

/* see in /home/epop/code/renew.c for the old function renew */
/* (may be useful for phonons) */

void add_particle(double tsim, int ng) {     // ng = 1 or nxg
  nelec++;
  if (nelec > maxnelec) {
	fprintf(stderr,"*** ERROR:  Too many particles (%d) ***\n",nelec);
	write_output_distro("final.txt");
	exit(-1);
  }
  if (ng == 1) {
	create_electron(nelec,2,T);
	Elec[X][nelec] = Xgrid[1] + ranmc()*DAgrid[1];
	Gxp[nelec] = 1;  //locate(Xgrid,1,nxg-1,Elec[X][nelec]);
  } else {
	create_electron(nelec,-2,T);
	Elec[X][nelec] = Xgrid[nxg] - ranmc()*DAgrid[nxg];
	Gxp[nelec] = nxg-1;
  }
  Elec[TS][nelec] = tsim - log(ranmc())/MAXSCAT;
  return;
}

void delete_particle(int i) {
  int II;
  for (II = 1; II <= TS; II++) {
	Elec[II][i] = Elec[II][nelec];   // copy over it
	iv[i] = iv[nelec];
	Gxp[i] = Gxp[nelec];
#ifdef NOISE
	memX[i] = memX[nelec];
#endif
  }
  nelec--;
  return;
}

void renew_contacts(double tsim) {
  // Need NSPgrid in super-electron numbers (w/out realsimscale)
  // from mcloop.  so move scaling by realsimscale inside poisson1()
  //
  // The stuff below only works in 1D right now (as does Poisson)
  int II, ndel, nadd, i, idel, iadd;
  double NLSP = NSPgrid[1];
  double NRSP = NSPgrid[nxg];

  /*for (i = 1; i <= nelec; i++) {
	if (Elec[X][i] < XLCONT) NLSP++;
	else if (Elec[X][i] > XRCONT) NRSP++;
	}*/

  /**************************************************************
   * LEFT contact
   **************************************************************/
  i=1; idel=0; iadd=0;
  /********* DELETE particles if too many **********/
  if (NLSP > NLCONT) {
	ndel = iround(NLSP - NLCONT);
	if (ndel > nelec)
	  nrerror("*** ERROR:  Too many SPs to delete at LEFT contact ***");
	do {   	  // delete (force exit) ndel particles
	  if (Elec[X][i] < XLCONT) {    // or locate based on Gxp[i]==1
		delete_particle(i);
		idel++;
	  }
	  i++;
	} while (idel != ndel);
	
	/********* ADD particles if too few **********/
  } else if (NLSP < NLCONT) {
	nadd = iround(NLCONT - NLSP);
	do {      // add (force inject) nadd particles
	  add_particle(tsim,1);
	  iadd++;
	} while (iadd < nadd);          // this also takes care of nadd = 0
  }

  /**************************************************************
   * RIGHT contact
   **************************************************************/
  i=1; idel=0; iadd=0;
  /********* DELETE particles if too many **********/
  if (NRSP > NRCONT) {
	ndel = iround(NRSP - NRCONT);
	if (ndel > nelec)
	  nrerror("*** ERROR:  Too many SPs to delete at RIGHT contact ***");
	do {   	  // delete (force exit) ndel particles
	  if (Elec[X][i] > XRCONT) {    // or locate based on Gxp[i]==1
		delete_particle(i);
		idel++;
	  }
	  i++;
	} while (idel != ndel);
	
	/********* ADD particles if too few **********/
  } else if (NRSP < NRCONT) {
	nadd = iround(NRCONT - NRSP);
	do {      // add (force inject) nadd particles
	  add_particle(tsim,nxg);
	  iadd++;
	} while (iadd < nadd);          // this also takes care of nadd = 0
  }

  return;
}
