/* these are column indices to track
   super-particle data (Elec matrix) */
#define X  1
#define Y  2
#define VX 3
#define VY 4
#define PX 5
#define PY 6
#define PZ 7
#define EE 8
#define TS 9

/* these are column indices to track
   grid-averaging in 1-D (GridAvg matrix) */
#define NEg 1    // number of electrons   //
#define EEg 2    // electron energy       //
#define VXg 3    // velocity              // all are AVERAGES
#define POg 4    // potential             //
#define FFg 5    // electric field        //

/* Set this for NOISE calculations in 1D, then recompile.  It speeds
   up the code a bit by not calculating per-grid averages and it
   calculates the Ramo theorem output every SUBHIST*DT seconds */
//#define NOISE

/* Choose one of three random number generators (see random.c) */
#define RAN1
//#define RANMT
//#define RAN48

/* Only use this for the 2-D raised SOI device for now */
//#define DUMBINIT

/* Use this to enable the charge-neutral boundaries */
//#define RENEW

/* Choose one of three sets of deformation potentials (see definitions.h) */
//#define CANALI
//#define BRUNETTI
#define POP

//#ifdef POP
//#define BZCLIP 1
//#else
#define BZCLIP 0
//#endif

#define ABS (-1.0)
#define EMS  1.0

#define LA 1
#define TA 2
#define LO 3
#define TO 4

#define MAXEF 100
#define MAXTSTEPS 1000
#define MAXLINE 500
#define MAXWORD 50
#define AVGSTEP 4

#ifndef NOISE
#define MAXHIST 50000     // enough for 1 fs steps up to 50 ps
#endif

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))
#define NR_END 1
#define FREE_ARG char*
#define AA(i,j) (*(AA + 2*(i) + (j) - 1))
#define UD(i,j) (*(UD + 2*(i) + (j) - 1))

// Universal Constants from http://physics.nist.gov/cuu/Constants
#define pi   3.14159265
#define pi2  6.28318531
#define echarge   -1.0             // electron charge 
#define ecoulom   -1.60217653e-19  // electron charge in Coulombs 
#define kb     8.61734318e-05      // eV/K 
#define hbar   6.58211915e-16      // eV*s 
#define m0     5.68562975e-16      // eV*s^2/cm^2
#define eps0   5.52634972e+05      // e/V/cm permittivity of free space

// material constants for silicon
#define asi    5.431e-08        // lattice constant, cm 
#define rho    1.4516e+12       // density (eV*s^2/cm^5) (2.329 g/cm^3) 
#define epssi  6.46582917e+06   // e/V/cm permittivity of Si 11.7*eps0

#define NPROC   18           // number of defined scattering mechanisms 
#define Zf       4.0         // f-scat degeneracy of final valley
#define Zg       1.0         // g-scat degeneracy of final valley 
#define vop      1.1e+05     // cm/s, optical phonon velocity 
#define tauOpAc  4.0e-12     // OP -> AC decay time

// global variables declared here 
#define BIN1   1000           // # bins along one dim >  nxg  
#define BIN2   20000          // # bins in two dim >  nxg*nyg 

// max allocation, used only in declaration of ScatMat, can
// go up to 10 eV in 1 meV steps -- or 5 eV in .5 meV steps
#define NEBINMAX   10000

//
// These are all the global variables that can change during execution
//

extern double XU, XD, DLA, DTA, ETAg, ELAg, ETAf, ELAf, ETOf, ELOg;
extern double DLOg, DLAg, DTAg, DTOf, DLAf, DTAf, EminTA, EminLA;
extern double QMIN, QMAX, PQMAX, Rws, QFp, QGp;
extern double Egap, ml, mt, mdos, alpha, Ni;
extern double T, MAXSCAT, DT, SIGEX, DESIGEX, DOPING, minDOPING, TMAX;
extern double VERROR, EFIELD, EFSTART, EFSTOP, EPROJX, EPROJY, EPROJZ;
extern int NEBIN, EFDIR, EFSTEPS, EFCOMMA, LOGSTEPS, MOBCALC, FASTSCAT;
extern int nphon, nxg, nyg, DIM, STATS, SUBHIST, nelec, maxnelec;
extern int POISSON, READRATES;
extern long seed;
extern double XMIN, XMAX, YMIN, YMAX, CHTOP, CHBOT;
extern double EMIN, EMAX, Elost;
extern double TTOT, TTRANS, SURFSCAT, DE, MOBILITY;
extern char INPUTFILE[MAXWORD], OUTPUTFILE[MAXWORD];
extern char FIELDFILE[MAXWORD], OUTDIR[MAXWORD];

extern int CLEFT, CRITE, isf, ioe, ioa, iae, iaa, itot;
extern double EphLA, EphTA, EphLO, EphTO;
extern int *iv, *Gxp, *Gyp, *GE;
extern double **Elec, *lambda, *MScat, *EF, *VF;
extern double *edist, *Ehist, *Vhist, *wcic, *rate, dxP, dyP, randummy;
extern double VXMAX, VYMAX;
extern double VLIM;              // max velocity likely to be encountered
extern int GLIM;                 // how many bins left or right to search
extern double *Ogrid, *Agrid, *NOgrid, *NAgrid, *ebins;
extern double *rateint, *Xgrid, *Ygrid, *DAgrid, *memX;
extern double **ScatMat, *DXgrid, *DYgrid, *AA, *UD, *bb;
extern double *EFXgrid, *EFYgrid, *POTgrid, *CHGgrid, *DOPgrid, *TXgrid;
extern double *NSPgrid, *NEpois, *Egrid, *oAAgrid, **GridAvg;
extern double *VXgrid, *VYgrid, *NO, *dNO, *NA, *dNA;
extern double *mx, *my, *mz, *pvx, *pvy, *pvz;
extern double alpha, o2alpha, alpha4, alpha2;
extern double LTAO[4][3];

// contact-related quantities
extern double XLCONT, XRCONT, NLCONT, NRCONT, OXLEF, OXRIG;
extern double realsimscale, Vtimeavg;

// accounting phonon distributions
#define DQph (QMAX/50.)
// DEph * NBph >= Max phonon energy (70 meV)
#define DEph 2.5e-4
#define NBph 280
extern double *NELph0, *NETph0, *Eph0;
extern double *NQLph0, *NQTph0, *Qph0;

//
// Function declarations and prototyping
//

/*************************** from randomc.c ***************************/
/*#ifdef RAN48
  void (*init_ranmc)(long);
  double (*ranmc)();
  #else*/
#ifdef RANMT
void init_ranmc(unsigned long);
#else
void init_ranmc(long);
#endif
double ranmc(void);
//#endif
double gasdev(void);
/********************* from defaults.c and mcloop.c *******************/
void defaults(void);
void mcloop(void (*)(int,double),FILE*,double*);
/***************************** from auxfns.c **************************/
size_t strlcat(char*,const char*,size_t);
size_t strlcpy(char*,const char*,size_t);
int ifloor(double);
int iround(double);
void nrerror(char*);
double *vector(int,int);
double **matrix(int,int);
int *ivector(int,int);
void printvector(double*,int,int,char*,char*);
void printivector(int*,int,int,char*,char*);
void print2vectors(double*,double*,int,int,char*,char*);
void printmatrix(double**,int,int,int,int,char*,char*,int);
void printgridavgs(double);
double meanmatrixcol(double**,int,int,int);
double sumvector(double*,int,int);
double meanvector(double*,int,int);
double normvector(double*,int,int);
double stdvector(double*,int,int);
void zerovector(double*,int,int);
void free_vector(double*,int,int);
double get_gamma(int);
double energy(int);
double velocity(double,double,double);
double ran1(void);
int locate(double*,int,int,double);
int locate_nearest(double*,int,double);
void cloud_in_cell(int,int,int);
void update_avg(double*,int,int,double);
void phonon_stats(double,double,double*,double*,double*,double*,int,int);
/**************************** from scat.c *****************************/
double get_efield(int,int);
double DOSe(double);
void ratecalc(double);
void change_equiv_valley(int);
void change_noneq_valley(int);
int intra_scat(int,double,int,double,double*,double*);
int inter_scat_f(int,int,double,double*,double*);
int inter_scat_g(int,int,double,double*,double*);
void scatter(int);
void drift0_100(int,double);
void drift0_any(int,double);
void drift1(int,double);
void drift2(int,double);
void poisson1(double*,double);
#ifdef RENEW
void add_particle(double,int);
void delete_particle(int);
void renew_contacts(double);
#endif
void calcscatrate(void);
/************************** from initial.c ****************************/
void create_electron(int,int,double);
void read_input_distro(char*);
void write_output_distro(char*);
void initial_distro(void);
int nocomment(char*,int);
void readinput(char*);
void echoinput(void);
//int boundary_condition(int);
int boundary_condition(int,double,double,double,double);
int k_boundary_condition(int);
/************************** from acrate3d.c ****************************/
double trapz(double*,double*,int,int);
double get_wq(double,int);
double get_nq(double,int);
double get_Mq(double,int,double);
double fabs_cost(double,double,double,int,double);
double find_root(double,double,double,double,int,double);
double acrate3(double,int,double);
