/*
  This function's single purpose is to be called inside
  input() and to set some initial default values for global
  variables declared in constants.h
*/
#include <string.h>

void defaults(void) {
  int i;
  double vla = LTAO[LA-1][1];          // LA phonon velocity
  double vta = LTAO[TA-1][1];          // TA phonon velocity

  EminLA = 0.5*mdos*vla*vla;
  EminTA = 0.5*mdos*vta*vta;

  // define these consistently with constants pi, asi, hbar
  QMIN = 1.0;          // nearly zero compared to QMAX
  QMAX = 2*pi/asi;     // 1.1569113068e+08
  PQMAX = hbar*QMAX;   // 7.6149280673e-08
  // Wigner-Seitz cell radius
  Rws = asi*pow(3./(16.*pi), 1./3.);   // 2.1224e-8 cm
  // f- and g-phonon norm, to be multiplied by QMAX
  QFp = sqrt(1. + 2.*0.15*0.15);       // 1.022252415
  QGp = 0.3;

#ifdef CANALI
  ETAf = 210.*kb;
  ELAf = 500.*kb;
  ETOf = 630.*kb;
  ETAg = 140.*kb;
  ELAg = 210.*kb;
  ELOg = 700.*kb;
#elif defined(BRUNETTI)
  ETAf = 220.*kb;
  ELAf = 550.*kb;
  ETOf = 685.*kb;
  ETAg = 140.*kb;
  ELAg = 215.*kb;
  ELOg = 720.*kb;
#elif defined(POP)
  // Fphon = 2-norm([1 .15 .15]) in constants.h
  ETAf = hbar*get_wq(QFp*QMAX,TA);     // ~ 20 meV
  // this one is actually on the LO branch here b/c Fphon > 1
  ELAf = hbar*get_wq(QFp*QMAX,LA);     // ~ 50 meV
  // see how get_wq() is implemented for LA when q > QMAX
  ETOf = hbar*get_wq(QFp*QMAX,TO);     // ~ 59 meV
  ETAg = hbar*get_wq(QGp*QMAX,TA);     // ~ 12 meV
  ELAg = hbar*get_wq(QGp*QMAX,LA);     // ~ 18 meV
  ELOg = hbar*get_wq(QGp*QMAX,LO);     // ~ 63 meV
  /* printf("LAmax = %g\nLOmin = %g\n",
		 hbar*get_wq(QMAX,LA), hbar*get_wq(QMAX,LO));
		 exit(0); */
#endif

  for (i = 1; i <= 6; i++) {
	if ((i+1)/2 == 1) {               // i = 1 or 2
	  mx[i] = ml; my[i] = mt; mz[i] = mt;
	  pvx[i] = ((double) i*2.-3.)*0.85*PQMAX;
	  // other components of pvx/pvy/pvz=0 when initialized w/vector()
	} else if ((i+1)/2 == 2) {        // i = 3 or 4
	  mx[i] = mt; my[i] = ml; mz[i] = mt;
	  pvy[i] = ((double) i*2.-7.)*0.85*PQMAX;
	} else {                          // i = 5 or 6
	  mx[i] = mt; my[i] = mt; mz[i] = ml;
	  pvz[i] = ((double) i*2.-11.)*0.85*PQMAX;
	}
  }

  return;
}
