/*
  this is called at the top of main.c and contains the
  DEFINITIONS of all global variable declared in constants.h
*/

/******************** deformation potentials, etc *********************/

#ifdef CANALI            // Canali's original set

double DTAf = 0.15e8;
double DLAf = 3.4e8;
double DTOf = 4.0e8;
double DTAg = 0.5e8;
double DLAg = 0.8e8;
double DLOg = 3.0e8;
double DLA = 9.0;
double DTA = 0.0;

#elif defined(BRUNETTI)  // Brunetti + Jacoboni's set

double DTAf = 0.3e8;     // 0.15 -- 0.3 e8
double DLAf = 2.0e8;     //---\__ these two are probably EQUAL
double DTOf = 2.0e8;     //---/   b/c they're so close in energy
double DTAg = 0.5e8;     // fixed
double DLAg = 0.8e8;     // fixed
double DLOg = 11.0e8;    // 3.0 -- 11.0 e8
double DLA = 9.0;
double DTA = 0.0;

#elif defined(POP)

double DTAf = 0.6e8;   // 18 meV
double DLAf = 3.5e8;   // 50 meV
double DTOf = 1.5e8;   // 57 meV
double DTAg = 0.7e8;   // 10 meV
double DLAg = 1.0e8;   // 19 meV
double DLOg = 6.0e8;   // 63 meV
double XU = 6.8;
double XD = 1.1;
double DLA = 6.39;
double DTA = 3.01;

#endif

#ifdef POP
// quadratic phonon dispersion parameters
double LTAO[4][3] = {{-0.196e-2,  9.00e5, 0.},          // LA
					 {-0.250e-2,  5.33e5, 0.},          // TA
					 {-0.156e-2,  0.0000, 9.87682e13},  // LO
					 { 0.137e-2, -2.91e5, 1.027e14}};   // TO
// this set is closer to matching TA/TO zone edge velocity to zero
/*double LTAO[4][3] = {{-0.200e-2,  9.01e5, 0.},        // LA
					 {-0.260e-2,  5.23e5, 0.},          // TA
					 {-0.160e-2,  0.0000, 9.88e13},     // LO
					 { 0.111e-2, -2.57e5, 1.02e14}};    // TO
*/
#else
double LTAO[4][3] = {{0.,  9.00e5, 0.},
					 {0.,  5.33e5, 0.},
					 {-0.156e-2,  0.0000, 9.87682e13},
					 { 0.137e-2, -2.91e5, 1.027e14}};
#endif

double EFIELD, ETAg, ELAg, ETAf, ELAf, ELOg, ETOf;
double Egap, ml, mt, mdos, alpha, Ni, EminTA, EminLA;
double QMIN, QMAX, PQMAX, Rws, QFp, QGp;

int NEBIN = 4000;
double T = 300.0;
double MAXSCAT = 1.0e10;  // 1/s maximum scattering rate, keep it low here
double DT = 5.e-15;       // should be < 1/MAXSCAT only if we solve Poisson
double SIGEX = 0.0;
double DESIGEX = 0.0;
double TMAX = 0.0;
double DOPING = 10.0;       // absurdly small to begin with so it's << Ni
double minDOPING = 1.0e30;  // absurdly large to begin with
double VERROR = 0.01;
double EPROJX = 1.0;
double EPROJY = 0.0;
double EPROJZ = 0.0;
double EFSTART = 1000.;
double EFSTOP = 0.;
int EFSTEPS = 0;
int EFDIR = 111;
int EFCOMMA = 0;
int LOGSTEPS = 0;
int MOBCALC = 0;
int FASTSCAT = 0;
int SUBHIST = 50;
int nelec = 10000;
int maxnelec = 10000;
int nphon = 0;
int nxg = 1;
int nyg = 1;
int DIM = 0;
int POISSON = 0;
int READRATES = 0;
int STATS = 0;
long seed = 0;
double XMIN = 0.0;
double XMAX = 1.e-4;       // cm 
double YMIN = 0.0;
double CHTOP = 0.0;
double YMAX = 1.e-4;
double CHBOT = 1.e-4;
double EMIN = 1.e-10;      // don't make comparable to DBL_EPSILON=2.22e-16
double EMAX = 1.9;         // eV
double DE = 0.5e-3;        // eV 
double Elost = 0.0;        // energy lost due to capping E < EMAX
double TTOT = 5.0e-12;
double TTRANS = 1000.0;     // absurdly large value (1000 seconds!!!)
double SURFSCAT = 0.0;
double MOBILITY = 0;
  
int CLEFT=0;
int CRITE=0;
int isf=0;
int ioe=0;
int ioa=0;
int iae=0;
int iaa=0;
int itot=0;

double EphLA = 0.0;      // total net (em-abs) energy emitted
double EphTA = 0.0;      // per phonon branch
double EphLO = 0.0;
double EphTO = 0.0;
double VXMAX=1.0e7;      // cm/s
double VYMAX=1.0e7;
double VLIM =6.0e7;      // haven't seen velocities larger than this value
int GLIM=2;              // how many bins left or right to search

// set contacts (for 2D) here for now 
double XLCONT, XRCONT, NLCONT, NRCONT;
double OXLEF, OXRIG;
  
// scaling factor 
double realsimscale = 1.;

// time-average velocity used in mcloop.c
double Vtimeavg = 0.;

char INPUTFILE[MAXWORD], OUTPUTFILE[MAXWORD];
char FIELDFILE[MAXWORD], OUTDIR[MAXWORD];

int *iv, *Gxp, *Gyp, *GE;
double **Elec, *lambda, *MScat, *EF, *VF;
double *edist, *wcic, *rate, dxP, dyP, randummy;
#ifndef NOISE
double *Ehist, *Vhist;
#endif
double *Ogrid, *Agrid, *NOgrid, *NAgrid, *ebins;
double *rateint, *Xgrid, *Ygrid, *DAgrid, *memX;
double *DXgrid, *DYgrid, *AA, *UD, *bb;
double **ScatMat;
double *EFXgrid, *EFYgrid, *POTgrid, *CHGgrid, *DOPgrid, *TXgrid;
double *NSPgrid, *NEpois, *Egrid, *oAAgrid, **GridAvg;
double *VXgrid, *VYgrid, *NO, *dNO, *NA, *dNA;
double *mx, *my, *mz, *pvx, *pvy, *pvz;
double *NELph0, *NETph0, *Eph0;
double *NQLph0, *NQTph0, *Qph0;
double o2alpha, alpha4, alpha2;
