#include <stdio.h>
#include "constants.h"
#include "boundary.c"
#include <math.h>

/******** return electric field (along x if ij=1, else along y) *********/
double get_efield(int j, int ij) {
  int iex, iey=1, iex2;
  double ef, *fld;

  if (DIM == 1) {
	iex = Gxp[j];
	cloud_in_cell(j, iex, iey);
	ef = wcic[1]*EFXgrid[iex]+wcic[2]*EFXgrid[iex+1];
	return ef;
  } else {
	iex = Gxp[j];
	iey = Gyp[j];
	cloud_in_cell(j, iex, iey);
	iex2 = (iex-1)*nyg+iey;
	if (ij == 1) {
	  fld = EFXgrid;
	} else {
	  fld = EFYgrid;
	}
	ef = wcic[1]*fld[iex2]  + wcic[2]*fld[iex2+1] +
	  wcic[3]*fld[iex2+nyg] + wcic[4]*fld[iex2+nyg+1];
	return ef;
  }
}

__inline void drift0_100(int j, double tau) {
  Elec[PX][j] -= EFIELD*tau;
  if (BZCLIP == 1) k_boundary_condition(j);
  Elec[EE][j] = energy(j);
  Elec[VX][j] = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j]);
}

__inline void drift0_any(int j, double tau) {
  Elec[PX][j] -= EPROJX*EFIELD*tau;
  Elec[PY][j] -= EPROJY*EFIELD*tau;
  Elec[PZ][j] -= EPROJZ*EFIELD*tau;
  if (BZCLIP == 1) k_boundary_condition(j);
  Elec[EE][j] = energy(j);
  Elec[VX][j] = EPROJX*velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j])
	+ EPROJY*velocity(my[iv[j]],Elec[PY][j]-pvy[iv[j]],Elec[EE][j])
	+ EPROJZ*velocity(mz[iv[j]],Elec[PZ][j]-pvz[iv[j]],Elec[EE][j]);
}

__inline void drift1(int j, double tau) {
  double xold, vxold, eenew;
  int bcgp;

  /******** store old velocity, need it for averaging below *********/
  vxold = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j]);
  /************** time-advance PX (recall p = hbar*k) ***************/
  /* may need to get avg efield, distance may be 5e-7*1e-15 = .5 nm */
  /************* implicitly assume echarge = -1.0 below *************/
  Elec[PX][j] -= get_efield(j,1)*tau;
  if (BZCLIP == 1) k_boundary_condition(j);
  /************ calculate new energy, velocity, position ************/
  Elec[EE][j] = energy(j);
  Elec[VX][j] = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j]);

  xold = Elec[X][j];
  Elec[X][j] += 0.5*tau*(vxold + Elec[VX][j]);
  /**** apply boundary condition, make sure particles stay inside ***/
  bcgp = boundary_condition(j,xold,0.0,vxold,0.0);
  /*if (bcgp == 0) {
	fprintf(stderr, "bcgp = %i  px = %g  x = %g  vx = %g\n",
	bcgp, Elec[PX][j], Elec[X][j], Elec[VX][j]);
	}*/
  /*************** update position-bin grid pointers ****************/
  if (bcgp) {
	Gxp[j] = locate(Xgrid, max(1,Gxp[j]-GLIM), min(nxg-1,Gxp[j]+GLIM), Elec[X][j]);
  }
}

/*********************************************************************/
void drift2(int j, double tau) {
  double xold, yold, vxold, vyold, eenew;
  int bcgp;

  vxold = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],Elec[EE][j]);
  Elec[PX][j] -= get_efield(j,1)*tau;

  vyold = velocity(my[iv[j]],Elec[PY][j]-pvy[iv[j]],Elec[EE][j]);
  Elec[PY][j] -= get_efield(j,2)*tau;

  if (BZCLIP == 1) k_boundary_condition(j);
  /********************* compute new energy **************************/
  /* may not be neccesary to do this -- maybe just compute gamma, use
	 it below and compute energy only at the top of scatter()        */
  eenew = energy(j);

  /*if (eenew > EMAX) {
	fprintf(stderr, "i=%i, Eold=%g, Enew=%g\n", j, Elec[EE][j], eenew);
	fprintf(stderr, "x=%g, y=%g (nm)\n", Elec[X][j]*1e7, Elec[Y][j]*1e7);
	}*/

  /******************* compute new velocities ************************/
  Elec[VX][j] = velocity(mx[iv[j]],Elec[PX][j]-pvx[iv[j]],eenew);
  Elec[VY][j] = velocity(my[iv[j]],Elec[PY][j]-pvy[iv[j]],eenew);

  /********* distance traveled dx = <v>*tau = tau*(vold+v)/2 *********/
  xold = Elec[X][j];
  Elec[X][j] += 0.5*tau*(vxold + Elec[VX][j]);
  yold = Elec[Y][j];
  Elec[Y][j] += 0.5*tau*(vyold + Elec[VY][j]);
  
  /**** apply boundary condition, make sure particles stay inside ***/
  //some time i should get around to doing the time-reversal shtuff...
  Elec[EE][j] = eenew;
  bcgp = boundary_condition(j,xold,yold,vxold,vyold);
  
  /*************** update position-bin grid pointers ****************/
  if (bcgp) {
	Gxp[j] = locate(Xgrid, max(1,Gxp[j]-GLIM), min(nxg-1,Gxp[j]+GLIM), Elec[X][j]);
	Gyp[j] = locate(Ygrid, max(1,Gyp[j]-GLIM), min(nyg-1,Gyp[j]+GLIM), Elec[Y][j]);
  }
}

void poisson1(double *u, double tsim) {
  int i;
  // AA is the tridiagonal operator matrix evaluated during initial.c
  for (i = 1; i <= nxg; i++) {
	// convert from super-particle numbers to electron density
	NEpois[i] = realsimscale*NSPgrid[i]/DAgrid[i];
	// this now gives actual charge in electrons per cm^3
#ifndef RENEW
	// this is one (bad?) way to force charge neutrality at the contacts
	//if ((i <= 1) || (i >= nxg)) NEpois[i] = DOPgrid[i];
#endif
	bb[i] = (DOPgrid[i]-NEpois[i])*DAgrid[i];       // *e=1

	// vince:
	//bb[i] = (sin(2.*pi*Xgrid[i]/Xgrid[nxg]) + Xgrid[i]/Xgrid[nxg])*1e19;
  }
  //printf("bb2 = %g\nbbn-1 = %g\n", bb[2], bb[nxg-1]);

  // boundary conditions
  bb[2]     = bb[2] - AA(2,1)*POTgrid[1];
  bb[nxg-1] = bb[nxg-1] - AA(nxg-1,nxg)*POTgrid[nxg];
  // NOTE: u is POTgrid from the way poisson1() is called

  // we're solving AA.u=b (UD.u=b) via tridiagonal elimination
  for (i=3; i <= nxg-1; i++) {
	bb[i] = bb[i] - bb[i-1]*AA(i,i-1)/UD(i-1,i-1);
  }

  // back-substitution to get the rest of the potentials 
  u[nxg-1] = bb[nxg-1]/UD(nxg-1,nxg-1);
  for (i = nxg-2; i >= 2; i--) {
	u[i] = (bb[i] - UD(i,i+1)*u[i+1])/UD(i,i);
  }
  // NOTE: u is POTgrid from the way poisson1() is called

  // now evaluate the electric fields at each non-uniform grid node
  // i should probably IMPOSE that the grid be UNIFORM as per discussion
  // with Ravaioli and the false self-forces that may arise from CIC
  for (i = 2; i <= nxg-1; i++) {
	// from initial.c DXgrid[i] = Xgrid[i+1] - Xgrid[i]
	/*EFXgrid[i] = -(-DXgrid[i]/DXgrid[i-1]/(DXgrid[i-1]+DXgrid[i])*u[i-1] +
	  (1./DXgrid[i-1]-1./DXgrid[i])*u[i] +
	  DXgrid[i-1]/DXgrid[i]/(DXgrid[i-1]+DXgrid[i])*u[i+1]);*/
	EFXgrid[i] = -(u[i+1]-u[i-1])/(Xgrid[i+1]-Xgrid[i-1]);
  }
  //EFXgrid[1] = -(u[2]-u[1])/DXgrid[1];
  //EFXgrid[nxg] = -(u[nxg]-u[nxg-1])/DXgrid[nxg-1];
  EFXgrid[1] = -(-3.*u[1]+4.*u[2]-u[3])/(Xgrid[3]-Xgrid[1]);
  EFXgrid[nxg] = -(3.*u[nxg]-4.*u[nxg-1]+u[nxg-2])/(Xgrid[nxg]-Xgrid[nxg-2]);

  //printvector(EFXgrid,1,nxg,".","ef.txt");
  //printvector(u,1,nxg,".","po.txt");
}
