#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "constants.h"

/* lets me read the line unless it's commented out with % or # */
int nocomment(char *line, int maxl) {
  int i = 0;
  while (isspace(line[i]) && (i <= maxl)) i++;    /* skip spaces & tabs */
  if (i > maxl)	return 0;
  else return ((line[i] !='%') && (line[i] !='#'));
}

void readinput(char *infile) {
  int ne=0;
  char line[MAXLINE], tok1[MAXWORD], *st;
  FILE *fp;

  setlinebuf(stdout);

  strcpy(INPUTFILE, "");
  strcpy(OUTPUTFILE, "");
  strcpy(FIELDFILE, "");

  /* open the raw data file if possible */
  fp = fopen(infile,"r");
  if (fp == NULL) {
	printf("\n****** WARNING:  could not open input file (%s) ******\n", infile);
	printf("************** Proceeding with default values **************\n\n");
	return;
  } else {
	printf("******** READING INPUT PARAMETERS FROM FILE %s ********\n\n",
		   infile);
  }

  CHTOP = YMIN;
  CHBOT = YMAX;
  while (fgets(line, MAXLINE-1, fp) != NULL) {
	/* read only if the line isn't a comment as determined by nocomment() */
	if (nocomment(line, MAXLINE)) {
	  strlcpy(tok1, strtok(line, "\t= "), MAXWORD);
	  if (strcmp(tok1, "TEMP") == 0) {
		T = atof(strtok(NULL, "\n\t= "));
		if ((T <= 0.) || (T >= 1000.)) T = 300.;
		// set some temperature-dependent quantities
		Egap = 1.1756 - 8.8131e-05*T - 2.6814e-07*T*T;
		alpha = 0.5*1.1250281/Egap;
	  } else if (strcmp(tok1,"DT") == 0) {
		DT = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1,"SIGEX") == 0) {
		SIGEX = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1,"EFSTART") == 0) {
		st = strtok(NULL, "\t=, ");
		while ((st != NULL) && (strstr(st,"#") == NULL) && (strstr(st,"%") == NULL)) {
		  ne++;
		  EF[ne] = atof(st);
		  st = strtok(NULL, "\t=, ");
		}
		EFSTART = EF[1];
		EFCOMMA = ne-1;
	  } else if (strcmp(tok1, "EFSTOP") == 0) {
		EFSTOP = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "EFSTEPS") == 0) {
		EFSTEPS = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "EFDIR") == 0) {
		EFDIR = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "LOGSTEPS") == 0) {
		LOGSTEPS = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "MOBCALC") == 0) {
		MOBCALC = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "VERROR") == 0) {
		VERROR = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DOPING") == 0) {
		DOPING = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "SUBHIST") == 0) {
		SUBHIST = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "NELEC") == 0) {
		nelec = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "TTRANS") == 0) {
		TTRANS = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "TTOT") == 0) {
		TTOT = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "SURFSCAT") == 0) {
		SURFSCAT = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "CHTOP") == 0) {
		CHTOP = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "CHBOT") == 0) {
		CHBOT = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "OXLEF") == 0) {
		OXLEF = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "OXRIG") == 0) {
		OXRIG = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "INPUTFILE") == 0) {
		strlcpy(INPUTFILE, strtok(NULL, "\n\t= "), MAXWORD);
	  } else if (strcmp(tok1, "OUTPUTFILE") == 0) {
		strlcpy(OUTPUTFILE, strtok(NULL, "\n\t= "), MAXWORD);
	  } else if (strcmp(tok1, "FIELDFILE") == 0) {
		strlcpy(FIELDFILE, strtok(NULL, "\n\t= "), MAXWORD);
	  } else if (strcmp(tok1, "EMAX") == 0) {
		EMAX = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "SEED") == 0) {
		seed = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DIM") == 0) {
		DIM = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "POISSON") == 0) {
		POISSON = atoi(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "FASTSCAT") == 0) {
		FASTSCAT = atoi(strtok(NULL, "\n\t= "));
#ifdef POP
	  } else if (strcmp(tok1, "XU") == 0) {
		XU = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "XD") == 0) {
		XD = atof(strtok(NULL, "\n\t= "));
#endif
	  } else if (strcmp(tok1, "DLOg") == 0) {
		DLOg = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DLAg") == 0) {
		DLAg = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DTAg") == 0) {
		DTAg = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DTOf") == 0) {
		DTOf = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DLAf") == 0) {
		DLAf = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "DTAf") == 0) {
		DTAf = atof(strtok(NULL, "\n\t= "));
	  } else if (strcmp(tok1, "ALPHA") == 0) {
		alpha = atof(strtok(NULL, "\n\t= "));
	  }
	}
  }

  fclose(fp);

  return;
}

void echoinput(void) {
  int i;
  double DEFIELD, efdirnorm;
  
  /************* sanity-check a few values **************/
  if (nelec < 0) nelec = 10000;
  if (FASTSCAT) FASTSCAT = 1;
  if ((DIM < 0) || (DIM > 2)) DIM = 0;
  if (LOGSTEPS) LOGSTEPS = 1;
  SIGEX = fabs(SIGEX);
  if (SIGEX > 1.) SIGEX = .1;
  alpha = fabs(alpha);
  if (alpha > 1.) alpha = 0.5;
  o2alpha = 1./(2.*alpha);
  alpha4 = 4.*alpha;
  alpha2 = 2.*alpha;

  /****** Set the few temperature-dependent quantities ******/
  Egap = 1.1756 - 8.8131e-05*T - 2.6814e-07*T*T;
  ml = 0.9163*m0;               // 5.2097e-16 eV*s^2/cm^2
  mt = 0.1880*m0*1.1756/Egap;   // 1.1220e-16 eV*s^2/cm^2
  /******* Change the transverse mass under strain (Rieger & Vogl) ******/
  if (SIGEX > 0.) mt = mt*(1.0 + 0.052*SIGEX);    // 0.199m0 for SIGEX=0.3
  mdos = pow(ml*mt*mt, 1./3.);  // 1.8718e-16 eV*s^2/cm^2
  // should rewrite Ni so it's also f(Egap) for strained Si...
  Ni = 5.71e19*pow(T/300.0,2.365)*exp(-6733.0/T);

  /*** Set some IMPORTANT defaults (phonon energies, mi and pvi's) ***/
  defaults();

  /************* ok, now pretty-print the inputs ************/
  printf("TEMP = %3.f [K]               |   NELEC = %i\n", T, nelec);
  printf("DT = %.4e [s]          |   SUBHIST = %i [x DT]\n", DT, SUBHIST);
  if (TTRANS < TTOT) {
	printf("Transient = %.4e [s]   |   Total time = %.4e [s]\n", TTRANS, TTOT);
#ifndef NOISE
	if ((DT*((double) SUBHIST)*((double) MAXHIST)) < (TTOT-TTRANS)) {
	  nrerror("\n*** ERROR: Not enough internal storage for subhistories  ***\n*** Increase MAXHIST in constants.h and recompile.       ***\n");
	}
#endif
  } else {
	printf("Transient = AUTO             |   Total time = %.4e [s]\n", TTOT);
#ifndef NOISE
	if ((DT*((double) SUBHIST)*((double) MAXHIST)) < TTOT) {
	  nrerror("\n*** ERROR: Possibly not enough internal storage for subhistories ***\n*** Increase MAXHIST in constants.h and recompile.               ***\n");
	}
#endif
  }
  //printf("FASTSCAT = %i\n", FASTSCAT);
  printf("Dimension = %i                |   EMAX = %g [eV]\n", DIM, EMAX);
#ifdef NOISE
  if (DIM != 1) {
	nrerror("\n*** ERROR: you can only run NOISE simulations in 1D ***\n");
  }
#endif

  if (DE*NEBINMAX < EMAX) {
	nrerror("*** ERROR: not enough storage for energy range.  Please lower EMAX ***\n");
  } else {
	// calculate NEBIN, assign energy range bins
	NEBIN = 2 + ifloor((EMAX-EMIN)/DE);
	ebins[1] = EMIN;
	for (i = 2; i <= NEBIN-1; i++) {
	  ebins[i] = EMIN+((double) (i-1))*DE;
	}
	ebins[NEBIN] = EMAX;
  }
#ifdef POP
  printf("INTRA-scat: XU   = %.2f  XD   = %.2f [eV]\n", XU, XD);
  DLA = sqrt((XD*XD + XD*XU + XU*XU*3./8.)*pi/2.);
  DTA = sqrt(pi)*XU/4.;
  printf("            LA   = %.2f  TA   = %.2f [eV]\n", DLA, DTA);
#else
  printf("INTRA-scat: LA   = %.2f  TA   = %.2f [eV]\n", DLA, DTA);
#endif
  printf("INTER-scat: TA-g = %.1f [meV] ....... %.2f [1e8 eV/cm]\n", ETAg*1e3, DTAg*1e-8);
  printf("            LA-g = %.1f ............. %.2f\n", ELAg*1e3, DLAg*1e-8);
  printf("            LO-g = %.1f ............. %.2f\n", ELOg*1e3, DLOg*1e-8);
  printf("            TA-f = %.1f [meV] ....... %.2f [1e8 eV/cm]\n", ETAf*1e3, DTAf*1e-8);
  printf("          LA/O-f = %.1f ............. %.2f\n", ELAf*1e3, DLAf*1e-8);
  printf("            TO-f = %.1f ............. %.2f\n", ETOf*1e3, DTOf*1e-8);
  /* printf("Maximum energy EMAX = %g eV (sets MAXSCAT)\n", EMAX); */
  fprintf(stderr, "Alpha = %g [1/eV]", alpha);
  
  if (DIM == 0) {
	if (TTRANS > TTOT) {
	  if ((VERROR < 0.) || (VERROR > 0.1)) {
		printf("\n*** ERROR: please choose 0 < VERROR < 0.1 ***\n");
		exit(-1);
	  } else {
		fprintf(stderr, "           |   VERROR = %g\n", VERROR);
	  }
	} else {
	  fprintf(stderr, "\n");
	}
	if (DOPING >= 1e20) {
	  printf("\n*** WARNING: Monet only works for doping up to 1e20 ***\n");
	  printf("\n************ Reducing to DOPING = 1e20 ****************\n\n");
	  DOPING = 1e20;
	} else if (DOPING >= 1e12) {
	  printf("DOPING = %g\n", DOPING);
	}
	if (EFSTEPS < 0) {
	  nrerror("\n*** ERROR: need positive value of EFSTEPS ***\n");
	} else if (EFSTEPS >= MAXEF) {
	  nrerror("\n*** ERROR: please reduce number of EFSTEPS ***\n");
	}
	if ((fabs(EFSTART) > 1.e7) || (fabs(EFSTOP) > 1.e7)) {
	  printf("\n*** ERROR: need FIELD value less than 1.0E7 (V/cm) ***\n");
	  exit(-1);
	} else {
	  EPROJX = (double) floor(EFDIR/100);
	  EPROJY = (double) floor((EFDIR % 100)/10);
	  EPROJZ = (double) (EFDIR % 10);
	  //printf("%g %g %g\n", EPROJX, EPROJY, EPROJZ);
	  efdirnorm = sqrt(EPROJX*EPROJX+EPROJY*EPROJY+EPROJZ*EPROJZ);
	  EPROJX = EPROJX/efdirnorm;
	  EPROJY = EPROJY/efdirnorm;
	  EPROJZ = EPROJZ/efdirnorm;
	  //printf("%g %g %g\n", EPROJX, EPROJY, EPROJZ);
	  fprintf(stderr, "EFIELD (%i) = %g", EFDIR, EFSTART);
	  if (EFCOMMA >= 1) {
		EFSTEPS = EFCOMMA;
		if (EFSTEPS >= MAXEF) {
		  nrerror("\n*** ERROR: please reduce number of EFSTEPS ***\n");
		}
	  	EFSTOP = EF[EFCOMMA+1];
		for (i = 2; i <= EFCOMMA+1; i++) fprintf(stderr, ", %g", EF[i]);
	  } else {
		EF[1] = EFSTART;
		if (MOBCALC == 1) {
		  EFSTEPS = 4;
		  LOGSTEPS = 0;
		  EF[1] = EFSTART + 0.3*EFSTART;
		  EF[2] = EFSTART + 0.15*EFSTART;
		  EF[3] = EFSTART;
		  EF[4] = EFSTART - 0.15*EFSTART;
		  EF[5] = EFSTART - 0.3*EFSTART;
		} else if (EFSTEPS != 0) {
		  if (EFSTEPS >= MAXEF) {
			nrerror("\n*** ERROR: please reduce number of EFSTEPS ***\n");
		  }
		  if (LOGSTEPS) {
			DEFIELD = (log10(fabs(EFSTOP)) - log10(fabs(EFSTART)))
			  /((double) EFSTEPS);
			for (i = 1; i <= EFSTEPS; i++) {
			  EF[i+1] = pow(10.,log10(fabs(EFSTART))+((double) i)*DEFIELD)
				*EF[1]/fabs(EF[1]);
			  fprintf(stderr, ", %g", EF[i+1]);
			}
		  } else {
			DEFIELD = (EFSTOP-EFSTART)/((double) EFSTEPS);
			for (i = 1; i <= EFSTEPS; i++) {
			  EF[i+1] = EFSTART+((double) i)*DEFIELD;
			  fprintf(stderr, ", %g", EF[i+1]);
			}
		  }
		}
	  }
	  fprintf(stderr, " [V/cm]\n");
	  if (MOBCALC == 1) {
		fprintf(stderr, "Performing MOBILITY calculation in %d field steps.\n", EFSTEPS+1);
	  }
	}
	if (SIGEX > 0.) {
	  DESIGEX = 0.67*SIGEX;
	  printf("STRAINED SILICON on Si(1-x)Ge(x) with x = %.2f, Esplit = %.3f [eV]\n",
			 SIGEX, DESIGEX);
	  if (EFDIR != 100) {
		printf("\n*** ERROR:  Please select EFDIR = 100 in Strained Silicon! ***\n");
		exit(-1);
	  }
	}
  } else {
	EFSTEPS = 0;
	if (!strlen(FIELDFILE)) {
	  if (DIM == 1) strcpy(FIELDFILE,"efield1.txt");
	  if (DIM == 2) strcpy(FIELDFILE,"efield2.txt");
	}
	printf("\nFIELDFILE = %s\n", FIELDFILE);
	if (POISSON) {
	  POISSON = 1;
	  printf("Computing fields SELF-CONSISTENTLY with the POISSON equation.\n");
	}
	if (DIM == 2) {
	  if ((SURFSCAT < 0.) || (SURFSCAT > 1.)) {
		printf("\n*** ERROR:  Invalid value for SURFSCAT = %g ***\n", SURFSCAT);
		printf("*** Please enter a value between 0.0 and 1.0 ***\n");
		exit(-1);
	  } else {
		printf("SURFSCAT = %f\n", SURFSCAT);
	  }
	  printf("OXLEF = %g <-----> OXRIG = %g\n",OXLEF,OXRIG);
	  printf("CHTOP = %g <-----> CHBOT = %g\n",CHTOP,CHBOT);
	}
  }

  if (seed != 0) printf("\n*** Running in DEBUG mode, SEED = %li ***\n", seed);

  return;
}
