#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <string.h>

/* http://www.mkssoftware.com/docs/man1/sed.1.asp */
/* http://www.student.math.uwaterloo.ca/~cs342/documents/refvsptr.html */
/* sed `s/(x|y|px|py|pz|vx|vy|Ee|ts)\[i\]/Elec\[i\]\[\1\]/g' boundary.c > boundary.c.new */

#include "constants.h"   /* global (common) variables */
#include "definitions.h"
#include "defaults.c"

int main(int argc, char *argv[]) {
  FILE *fp;
  int i, j, ngrid, switches=0;
  char infile[MAXWORD], dirlogfile[MAXWORD], logfile[MAXWORD];
  char datestamp[MAXLINE], md[MAXLINE];
  void (*drift_ptr)(int,double);
  double dummy=0.0, tres, telapsed, begthist, verrinit, ttotinit;
  double mob=0.0, vsum=0.0, efsum=0.0, currentI, powerIV, powerPH;
  struct timeval startT, stopT;
  struct tm *time_struct;
  time_t timenull = time(NULL);
  time_struct = localtime(&timenull);
  strftime(datestamp, sizeof(datestamp), "-%d%b%g-%H%M%S", time_struct);

  mx = vector(1,6); my = vector(1,6); mz = vector(1,6);
  pvx = vector(1,6); pvy = vector(1,6); pvz = vector(1,6);

  EF = vector(1,MAXEF);  VF = vector(1,MAXEF);
#ifndef NOISE
  Ehist = vector(1,MAXHIST);  Vhist = vector(1,MAXHIST);
#endif
  Xgrid = vector(1,BIN1); DXgrid = vector(1,BIN1);
  AA = vector(1,3*BIN1); UD = vector(1,3*BIN1); bb = vector(1,BIN1);
  EFXgrid = vector(1,BIN2); EFYgrid = vector(1,BIN2); DOPgrid = vector(1,BIN2);
  TXgrid = vector(1,BIN1);
  POTgrid = vector(1,BIN2); DAgrid = vector(1,BIN2); Ogrid = vector(1,BIN2);
  GridAvg = matrix(BIN1,FFg);  // useful in 1D only, for now
  Agrid = vector(1,BIN2); NOgrid = vector(1,BIN2); NAgrid = vector(1,BIN2); 
  CHGgrid = vector(1,BIN2); VXgrid = vector(1,BIN2); Egrid = vector(1,BIN2);
  NSPgrid = vector(1,BIN2); NEpois = vector(1,BIN2);
  rate = vector(1,NPROC); wcic = vector(1,4);
  ebins = vector(1,NEBINMAX);  edist = vector(1,NEBINMAX);
  //ScatMat = scatmat_dummy;
  ScatMat = matrix(NEBINMAX,NPROC);

  strlcpy(OUTDIR, "out", MAXWORD);

  for (i = 0; i < argc; i++) {
	if (strncmp(argv[i],"-h",2) == 0) {
	  printf("Usage: monet [options] infile logfile\n");
	  printf("  -d        Append date stamp to output directory\n");
	  printf("  -h        Print this help\n");
	  printf("  -r        Read scattering rates from scatrates.txt\n");
	  printf("  infile    <monet.in>\n");
	  printf("  logfile   <monet.log>\n");
	  exit(0);
	} else if ((strncmp(argv[i],"-r",2) == 0) && (READRATES == 0)) {
	  READRATES = 1;
	  switches++;
	} else if (strncmp(argv[i],"-d",2) == 0) {
	  strlcat(OUTDIR, datestamp, MAXLINE);
	  switches++;
	}
  }

  if (argc-switches >= 2) strlcpy(infile, argv[switches+1], MAXWORD);
  else strlcpy(infile, "monet.in", MAXWORD);
  if (argc-switches >= 3) strlcpy(logfile, argv[switches+2], MAXWORD);
  else strlcpy(logfile, "monet.log", MAXWORD);

  // create the output directory
  snprintf(md, MAXLINE, "mkdir -p %s", OUTDIR);
  system(md);
  // store a copy of the input file in the output dir
  snprintf(md, MAXLINE, "cp -f %s %s", infile, OUTDIR);
  system(md);
  
  // read parameters from input file, or adopt defaults
  readinput(infile);
  // pretty-print parameters <--- this also calls defaults()
  echoinput();

#ifdef RENEW
  maxnelec = iround(1.2*nelec);
#else
  maxnelec = nelec;
#endif

  Elec = matrix(TS,maxnelec);
  iv = ivector(1,maxnelec);

  /*printf("Op-Ac border energy = %g or %g\n",
	hbar*get_wq(QMAX,LA), hbar*get_wq(QMAX,LO));*/

  //enable these if i start messing w/FASTSCAT again
  //GE = ivector(1,nelec); MScat = vector(1,nelec); lambda = vector(1,nelec);

  if (DIM == 0) {
	Eph0 = vector(1,NBph);
	NETph0 = vector(1,NBph); NELph0 = vector(1,NBph);
	for (i = 2; i <= NBph; i++) Eph0[i] = Eph0[i-1] + DEph;
  } else {
	Gxp = ivector(1,maxnelec);
#ifdef NOISE
	memX = vector(1,maxnelec);
#endif
	for (i = 1; i <= BIN1; i++) {
	  for (j = 1; j <= FFg; j++) GridAvg[i][j] = 0.0;
	}
	if (DIM == 2) {
	  Ygrid = vector(1,BIN1); DYgrid = vector(1,BIN1);
	  VYgrid = vector(1,BIN2); oAAgrid = vector(1,BIN2);
	  Gyp = ivector(1,maxnelec);
	}
  }

  /*********************** initialize random seeds **********************/
  if (seed == 0) {
	seed = -1;
	dummy = ranmc();                          // initialize w/seed < 0
	init_ranmc((long) timenull * getpid());
	//seed = (long) timenull * getpid();           // set seed ranmc
	// srandom(timenull * getpid());       //    set seed for ran1
  } else {
	init_ranmc(seed);
  }

  /****************  initial distributions (reads DOPING) ***************/
  initial_distro();

  /******* pre-calculate scat rates per energy bin (needs DOPING) *******/
  calcscatrate();

  /************* now initial particle times (needs MAXSCAT) *************/
  for (i = 1; i <= nelec; i++) {
	Elec[TS][i] = -log(ranmc())/MAXSCAT;
  }

  /************************** write stuff to log ************************/
  setlinebuf(stdout);
  snprintf(dirlogfile, MAXWORD, "%s/%s", OUTDIR, logfile);
  fp = fopen(dirlogfile, "w");
  setlinebuf(fp);
  fprintf(fp, "%% Temperature = %g K\n", T);
  fprintf(fp, "%% Time Step = %g sec\n", DT);

  printf("\n********************** START MONET RUN **********************\n");

  /**************** figure out which drift function to use **************/
  if (DIM == 0) {
	if (EFDIR == 100)
	  drift_ptr = drift0_100;
	else
	  drift_ptr = drift0_any;
  } else if (DIM == 1) {
	drift_ptr = drift1;
  } else {
	drift_ptr = drift2;
  }

  write_output_distro("initial.txt");

  /****************** start counting elapsed runtime ********************/
  gettimeofday(&startT, 0);

  /************* this runs the single case for 0-, 1- and 2-D ***********/
  ttotinit = TTOT; verrinit = VERROR;
  EFIELD = EF[1];
  do {
	mcloop(drift_ptr, fp, &begthist);
	if (STATS == 0) {
	  fprintf(stderr, "\nCouldn't converge within given constraints.\n");
	  TTRANS = 0.0;
	  VERROR = 2.*VERROR;
	  TTOT = 2.*TTOT;
	  fprintf(stderr,"Doubling VERROR to %.4f and TTOT to %.4e [sec] and trying again.\n\n", VERROR, TTOT);
	}
  } while (STATS == 0);

  /***** run more mcloops only in 0-D if several EFSTEPS requested *****/
  if ((DIM == 0) && (EFSTEPS != 0)) {
	if (MOBCALC) VF[1] = MOBILITY*fabs(EF[1]);

	for (i = 2; i <= EFSTEPS+1; i++) {
	  // reset velocity error and total time to original values
	  TTOT = ttotinit; VERROR = verrinit;
	  EFIELD = EF[i];
	  do {
		// run the monte carlo loop
		mcloop(drift_ptr, fp, &begthist);
		// adjust velocity error and total time if we didn't converge
		if (STATS == 0) {
		  fprintf(stderr, "\nCouldn't converge within given constraints.\n");
		  VERROR = 2.*VERROR;
		  TTOT = 2.*TTOT;
		  fprintf(stderr, "Doubling VERROR to %.4f and TTOT to %.4e [sec] and trying again.\n\n", VERROR, TTOT);
		}
	  }	while (STATS == 0);
	  if (MOBCALC) VF[i] = MOBILITY*fabs(EF[i]);
	}
  }

  /********************** get final elapsed time ************************/
  gettimeofday(&stopT, 0);
  telapsed = (double) (stopT.tv_sec - startT.tv_sec) +
	            1.e-6*(stopT.tv_usec - startT.tv_usec);

  /************** write outputs, compute final grid averages ************/
  printf("\n********************** PROGRAM OUTPUTS **********************\n");

  /**************** write output particle distribution *****************/
  if (strlen(OUTPUTFILE)) write_output_distro(OUTPUTFILE);

  printf("\n");

#ifndef NOISE  
  tres = TTOT - begthist;

  if (DIM) {
	printf("begthist = %g\n", begthist);
	currentI = fabs(realsimscale*ecoulom*((double) CRITE-CLEFT)/tres);
	powerIV  = fabs((POTgrid[nxg]-POTgrid[1])*currentI);
	printf("Current I = %.3e (A/cm^2)\n", currentI);
	printf("Power I*V = %.3e (W/cm^2)\n", powerIV);
	powerPH = realsimscale*ecoulom*(EphLO+EphTO+EphLA+EphTA)/tres;
	printf("Power Pho = %.3e (W/cm^2)\n\n", fabs(powerPH));
	printf("OpPower = %.3e [eV/s]           |", (EphLO+EphTO)/tres);
	printf("  AcPower = %.3e [eV/s]\n", (EphLA+EphTA)/tres);

	if (DIM == 1) ngrid = nxg;
	else ngrid = nxg*nyg;

	for (i = 1; i <= ngrid; i++) {
	  /* this entire mumbo-jumbo must be fixed when the
		 Poisson equation is solved, b/c NSPgrid is different */
	  if (NSPgrid[i] == 0) {
		Egrid[i] = 1.5*kb*T;
		VXgrid[i] = 0.;
		if (DIM == 2) VYgrid[i] = 0.;
	  } else {
		Egrid[i] = Egrid[i]/NSPgrid[i];
		VXgrid[i] = VXgrid[i]/NSPgrid[i];
		if (DIM == 2) VYgrid[i] = VYgrid[i]/NSPgrid[i];
	  }
	  /******** Super-Particle --> Electron Density conversion **********/
	  if (POISSON == 0) {
		NSPgrid[i] = realsimscale*NSPgrid[i]*DT/tres/DAgrid[i];
	  }
	  /******** these are per-grid phonon stats ********/
	  Agrid[i] = realsimscale*Agrid[i]/tres/DAgrid[i];
	  NAgrid[i] = realsimscale*NAgrid[i]/tres/DAgrid[i];
	  Ogrid[i] = realsimscale*Ogrid[i]/tres/DAgrid[i];
	  NOgrid[i] = realsimscale*NOgrid[i]/tres/DAgrid[i];
	}

	//printvector(Egrid, 1, ngrid, OUTDIR, "egrid.txt");
	//printvector(NSPgrid, 1, ngrid, OUTDIR, "ngrid.txt");
	//printvector(VXgrid, 1, ngrid, OUTDIR, "vxgrid.txt");
	//if (DIM == 2) printvector(VYgrid, 1, ngrid, OUTDIR, "vygrid.txt");
	//printvector(Agrid, 1, ngrid, OUTDIR, "agrid.txt");
	//printvector(Ogrid, 1, ngrid, OUTDIR, "ogrid.txt");
	//printvector(NAgrid, 1, ngrid, OUTDIR, "nagrid.txt");
	//printvector(NOgrid, 1, ngrid, OUTDIR, "nogrid.txt");
	//print2vectors(Xgrid,NOgrid, 1, ngrid, OUTDIR, "xnogrid.txt");
	//print2vectors(Xgrid,POTgrid, 1, ngrid, OUTDIR, "pot.txt");
	/*
	  if (DIM == 1) {
	  // convert electrons per grid into electron density
	  for (i = 1; i <= nxg; i++) {
		GridAvg[i][NEg] = realsimscale*GridAvg[i][NEg]/DAgrid[i];
	  }
	  // write out grid averages
	  printmatrix(GridAvg,1,nxg,1,FFg,OUTDIR,"avgs.txt",2);
	}
	*/
	if (POISSON) {
	  //print2vectors(Xgrid,NSPgrid,1,nxg,OUTDIR,"ne.txt");
	  //print2vectors(Xgrid,POTgrid,1,nxg,OUTDIR,"pot.txt");
	}
	// print2vectors(x,lambda,1,nelec,"lam.txt");

  } else if (DIM == 0) {
	if (EFSTEPS == 0) {
	  printf("Saving L/T phonon distributions (lphon.txt, tphon.txt)...\n");
	  for (i = 1; i <= NBph; i++) {
		NELph0[i] = NELph0[i]*realsimscale/tres;
		NETph0[i] = NETph0[i]*realsimscale/tres;
	  }
	  print2vectors(Eph0,NELph0,1,NBph,OUTDIR,"lphon.txt");
	  print2vectors(Eph0,NETph0,1,NBph,OUTDIR,"tphon.txt");
	  printf("Saving electron energy distribution to file (edist.txt)...\n");
	  print2vectors(ebins,edist,1,NEBIN,OUTDIR,"edist.txt");
	}
	if (MOBCALC) {
	  MOBILITY = fabs(sumvector(VF,1,EFSTEPS+1)/sumvector(EF,1,EFSTEPS+1));
	  printf("Mobility = %.0f [cm^2/V/s]\n",MOBILITY);
	  fprintf(fp,"Mobility = %.0f [cm^2/V/s]\n",MOBILITY);
	}
  }
#endif

  /*  itot = isf+ioe+ioa+iae+iaa;
  printf("%d self  %d opem  %d opabs  %d accem  %d accabs\n",
  isf,ioe,ioa,iae,iaa); */

  /*  printf("\nSummary:\n");
	  ((double) ioe)/tres, ((double) ioa)/tres, ((double) iae)/tres,
	  ((double) iaa)/tres); */

  if (DIM) {
	printf("One Simulated Electron = %g Real Electrons\n", realsimscale);
	printf("[assuming device dimensions not implicitly defined = 1 cm]\n");
	fprintf(fp, "\n%% One Simulated Electron = %g Real Electrons\n", realsimscale);
	fprintf(fp, "%% [assuming device dimensions not implicitly defined = 1 cm]\n");
  }

  printf("Total Simulation Time:  %g sec\n", telapsed);
  fprintf(fp,"%% Total Simulation Time:  %g sec\n", telapsed);
  if (EFSTEPS == 0) {
	printf("Time per Particle per psec:  %g milisec\n\n",
		   telapsed*1e3/((double) nelec)/(TTOT*1.e12));
	fprintf(fp,"%% Time per Particle per psec:  %g milisec\n\n",
			telapsed*1e3/((double) nelec)/(TTOT*1.e12));
  }

  fclose(fp);

  /* printf("VXMAX = %.4e   ", VXMAX);
	 if (DIM == 2) printf("VYMAX = %.4e", VYMAX);
	 printf("\n"); */

  /*
  printf("\n************************************************************\n");
  printf("\n   Starting Phonon BTE\n");
  phon1bte();
  */

  printf("Writing (again) Scattering Rates to File --> scatrates2.txt\n");
  printmatrix(ScatMat, 1, NEBIN, 1, NPROC, ".", "scatrates2.txt", 1);

  system("diff scatrates.txt scatrates2.txt");
  
  //printf("ScatMat = %p\n", ScatMat);
  //return 1;
}
