#include <stdio.h>
#include <math.h>
#include "constants.h"
#include "assert.h"

void mcloop(void (*drift_ptr)(int,double), FILE *fp, double *begthist) {
  int i, ih=1, istep=0, ihstart=1, ix=1, iy=1, DONE=0;
  double t1, tdt, tsim=0.0, Emean=0.0, Vmean=0.0, tres, powerPH;
  double Estd, Vstd, endthist = TTOT, Emeaninst, Vmeaninst, avgcorr=0.;
  double ihot=0.0, icold=0.0, ramo=0.0, w1, w2, Length, Length2, travel;
  double currentN, currentJ;
  char str[MAXWORD];
  FILE *out;
#ifdef NOISE
  double Ehistih, Vhistih;
#endif

  setlinebuf(stdout);

  STATS = 0;
  EphLA = 0.0; EphTA = 0.0; EphLO = 0.0; EphTO = 0.0;

#ifndef NOISE
  if (DIM == 0) fprintf(stderr, "EFIELD = %g [V/cm] ", EFIELD);
  // this sets the starting value of the averages (ih = 1)
  if ((EFSTEPS == 0) && (DIM == 0)) {
	Ehist[ih] = meanmatrixcol(Elec,1,nelec,EE);
	Vhist[ih] = meanmatrixcol(Elec,1,nelec,VX);
  }
  // otherwise Vhist[1] and Ehist[1] are memorized at the end of mcloop
#endif

  /********** start big MC loop for all particles and all time **********/
  while ((tsim <= TTOT) && (DONE == 0)) {

	tdt = tsim + DT;
	/************ drift and scatter for each particle *******************/
	for (i = 1; i <= nelec; i++) {
	  t1 = tsim;
	  while (Elec[TS][i] <= tdt) {
		drift_ptr(i, Elec[TS][i] - t1);
		scatter(i);
		t1 = Elec[TS][i];
		// if (FASTSCAT) ts(i) -= log(ranmc())/MScat[GE[i]]; else
		Elec[TS][i] -= log(ranmc())/MAXSCAT;
	  } 
	  drift_ptr(i, tdt-t1);
	}
	/********* end drift and scatter for each particle ******************/
	
	// completed one step and time advanced by DT
	istep++;
	tsim += DT;
	
	/*********** Quantities to compute with EACH time step DT ***********/

	// instantaneous averages
	Emeaninst = meanmatrixcol(Elec,1,nelec,EE);
	Vmeaninst = meanmatrixcol(Elec,1,nelec,VX);

	// cumulative averages over SUBHIST, divide them below
	Emean += Emeaninst;
	Vmean += Vmeaninst;

	// more CUMULATIVE averages per grid bin
	if (DIM) {
	  // zero out the electron density grid //if (POISSON)
	  zerovector(NSPgrid, 1, nxg);
	  zerovector(Egrid, 1, nxg);
	  zerovector(VXgrid, 1, nxg);
	  if (DIM == 2) zerovector(VYgrid, 1, nyg);
#ifdef RENEW
	  /****** maybe just move this part inside renew_contacts() ******/
	  if (DIM == 1) {
		// first, must count particles to know how much to renew
		for (i = 1; i <= nelec; i++) {
		  ix = Gxp[i];
		  cloud_in_cell(i, ix, iy);    // iy=1 b/c this is only in 1-D
		  // update the super-particle count
		  update_avg(NSPgrid, ix, iy, 1.0);
		}
		// now renew what's at the contacts
		renew_contacts(tsim);
	  }
	  zerovector(NSPgrid, 1, nxg);
#endif

	  for (i = 1; i <= nelec; i++) {
		// lambda[i] = hbar*2.*pi/(fabs(Elec[PX][i]));
		ix = Gxp[i];
		if (DIM == 2) iy = Gyp[i];   // in 1D iy=1 from above
		cloud_in_cell(i, ix, iy);
		// update the super-particle count
		update_avg(NSPgrid, ix, iy, 1.0);
		// update the particle energy distribution -- CUMULATIVE
		update_avg(Egrid, ix, iy, Elec[EE][i]);
		// update the particle velocity distribution -- CUMULATIVE
		update_avg(VXgrid, ix, iy, Elec[VX][i]);
		if (DIM == 2) update_avg(VYgrid, ix, iy, Elec[VY][i]);
	  }

	  if (POISSON) poisson1(POTgrid,tsim);

	  if (STATS && (DIM == 1)) {  // maybe collect GrdAvg here
		/* w1 = ((double) (ih-ihstart))/((double) (ih-ihstart+1));
		   w2 = 1.0/((double) (ih-ihstart+1)); */
		if ((tsim - (*begthist)) > 0.5*DT) {
		  w2 = DT/(tsim - (*begthist));
		} else {
		  w2 = 1.0;
		}
		w1 = 1.0 - w2;
		for (i = 1; i <= nxg; i++) {
		  GridAvg[i][NEg] = w1*GridAvg[i][NEg] + w2*NSPgrid[i];
		  if (NSPgrid[i] == 0) {
			GridAvg[i][EEg] = w1*GridAvg[i][EEg] + w2*1.5*kb*T;
			GridAvg[i][VXg] = w1*GridAvg[i][VXg] + w2*0.0;
		  } else {
			GridAvg[i][EEg] = w1*GridAvg[i][EEg] + w2*Egrid[i]/NSPgrid[i];
			GridAvg[i][VXg] = w1*GridAvg[i][VXg] + w2*VXgrid[i]/NSPgrid[i];
		  }
		  GridAvg[i][POg] = w1*GridAvg[i][POg] + w2*POTgrid[i];
		  GridAvg[i][FFg] = w1*GridAvg[i][FFg] + w2*EFXgrid[i];
		}
		//printf("w1 = %g ... w2 = %g\n", w1, w2);
	  }

	} else if ((DIM == 0) && (EFSTEPS == 0)) {
	  // in 0-D compute electron energy distribution only
	  for (i = 1; i <= nelec; i++) {
		edist[1 + ifloor(Elec[EE][i]/DE)] += 1.;
	  }
	}

	if ((TTRANS > DT*((double) SUBHIST)) && (istep == SUBHIST)) {
	  if (TTRANS > TTOT) {
		fprintf(stderr, "\nWaiting for transient (AUTO) ");
	  } else {
		fprintf(stderr, "\nWaiting for transient (%g s) ", TTRANS);
	  }
	}

	/***************************************************************
	  Every SUBHISTory (SUBHIST*DT seconds)
	  Figure out time-averages and what to print out
	***************************************************************/

	if ((istep % SUBHIST) == 0) {
	  ih++;
	  // compute averages over SUBHISTory
#ifdef NOISE
	  Ehistih = Emean/((double) (SUBHIST));
	  Vhistih = Vmean/((double) (SUBHIST));
#else
	  Ehist[ih] = Emean/((double) (SUBHIST));
	  Vhist[ih] = Vmean/((double) (SUBHIST));
#endif

	  // don't forget to set STATS back to 0 above once avgcorr sets it
	  if (STATS) {
#ifdef NOISE
		ramo = 0.0;
		Length  = XMAX-XMIN;
		Length2 = 0.5*Length;
		for (i = 1; i <= nelec; i++) {
		  travel = Elec[X][i] - memX[i];
		  if (travel < -Length2) travel += Length;
		  else if (travel > Length2) travel -= Length;
		  ramo += travel;
		  memX[i] = Elec[X][i];
		}
		ramo = -ecoulom*ramo/(XMAX-XMIN)/(((double) SUBHIST)*DT);
		fprintf(stderr," % .3e  % .5f  % .5e  % .5e\n",
				tsim,Ehistih,Vhistih,ramo);
		fprintf(fp," % .5e  % .5f  % .5e  % .5e\n",
				tsim,Ehistih,Vhistih,ramo);
#else
		if (DIM) {
		  tres = tsim-(*begthist);
		  currentN = (double) (CRITE-CLEFT);
		  if (tres == 0.0) {
			currentJ = 0.0;
		  } else {
			currentJ = fabs(realsimscale*ecoulom*currentN)/tres;
		  }
		  // in 1D or 2D stderr and logfile are the same
		  fprintf(stderr,"% .3e % .5f % .3e % .3e % .3e % .3e\n",
				  tsim,Ehist[ih],Vhist[ih],currentJ,EphLO+EphTO,EphLA+EphTA);
		  //% 5d %3.1f %3.1f\n");  ,nelec,NSPgrid[1],NSPgrid[nxg]);
		  fprintf(fp,    "% .3e % .5f % .3e % .3e % .3e % .3e\n",
				  tsim,Ehist[ih],Vhist[ih],currentJ,EphLO+EphTO,EphLA+EphTA);
		  if (DIM == 1) {
			// print out the grid averages every subhistory!
			printgridavgs(tsim-(*begthist));
			// print out the grid averages every subhistory!

			/* write out particle (x,E) for 1-D movie */
			/*snprintf(str, MAXWORD, "%s/monet%d.out",OUTDIR,istep/10);
			out = fopen(str,"w");
			for (i = 1; i <= nelec; i++) {
			  ix = Gxp[i];
			  cloud_in_cell(i, ix, iy);
			  fprintf(out,"% .5e % .5e % .5e\n",Elec[X][i],Elec[EE][i],
					  Elec[EE][i]-(wcic[1]*GridAvg[ix][POg]+wcic[2]*GridAvg[ix+1][POg]));
			}
			fclose(out);
			*/
			
		  }
		} else {
		  // 0-DIM
		  fprintf(stderr," % .3e  % .5f  % .4e  % .4e  % 2.4f\n",
				  tsim,Ehist[ih],Vhist[ih],Vtimeavg,avgcorr);
		  // the intermediate steps are commented out for multiple E-fields
		  if (EFSTEPS) fprintf(fp,"%%");
		  // write intermediate step (per-history) statistics
		  fprintf(fp," % .3e  % .5f  % .4e  % .4e  % .4e  % .4e  % .4e  % 2.4f\n",
				  tsim,Ehist[ih],fabs(Vhist[ih]),EphLA,EphTA,EphLO,EphTO,avgcorr);
		}

		// bail early in 0-D multi-steps if velocity error small enough
		if ((DIM == 0) && (ih > ihstart+2) && (EFSTEPS > 0)) {
		  Vmean = meanvector(Vhist,ihstart,ih);   // it's OK to reuse Vmean
		  Vstd = stdvector(Vhist,ihstart,ih);
		  if (fabs(Vstd/Vmean) < VERROR) DONE = 1;
		  endthist = tsim;
		}
		// but only bail early when doing multiple steps, b/c in a single
		// step i want the phonons stats to accumulate during all of TTOT
#endif
	  } else {
		fprintf(stderr, ".");
	  }

	  /*** Running average of the past AVGSTEP (=4 default) SUBHISTories ***/
	  if (((ih > AVGSTEP) && (DIM == 0) && (tsim > TTRANS))
		  || ((DIM) && (tsim > TTRANS))) {
		// if EFSTEPS >= 1 this should reuse Vtimeavg from previous run
#ifndef NOISE
		if ((ih > AVGSTEP) && (DIM == 0)) {
		  Vtimeavg = 0.;
		  for (i = ih-AVGSTEP; i < ih; i++) Vtimeavg += Vhist[i];
		  Vtimeavg = Vtimeavg/((double) AVGSTEP);
		  avgcorr = fabs(Vhist[ih]-Vtimeavg)/Vtimeavg;
		}
#endif
		/* begin STATS if Vhist gets within VERROR of time-avg velocity
		   OR if i've set TTRANS to something reasonable; note that in 1- or 2-D
		   avgcorr isn't computed so it's zero by default */
		if (((fabs(avgcorr) < VERROR) || (tsim > (TTRANS+0.5*DT))) && (STATS == 0)) {
		  STATS = 1;
		  *begthist = tsim;
		  ihstart = ih;
#ifdef NOISE
		  fprintf(stderr, "\n    t[s]       E[eV]     v[cm/s]       Noise[A]\n");
		  fprintf(fp, "%%    t[s]       <E>[eV]   <vx>[cm/s]     Noise[A]\n");
#else
		  fprintf(stderr, "\n   t[s]      E[eV]    v[cm/s]   J[A/cm^2]\n");
		  if (DIM) {
			fprintf(fp, "%%  t[s]      E[eV]    v[cm/s]   J[A/cm^2]    EphOp      EphAc\n");
		  } else {
			fprintf(fp, "%%  t[s]      E[eV]     v[cm/s]\n");
		  }
#endif
		}
	  }

	  // reset averages over SUBHIST
	  Emean = 0.;
	  Vmean = 0.;
	}
	/**************** end collecting stats for this DT ******************/

	/********* end big MC loop for all particles and all time ***********/
  }


#ifndef NOISE

  // this makes sure we've got at least 3 decent data points
  if ((DIM == 0) && (ih-ihstart <= 2)) STATS = 0;
  // otherwise it sets STATS to 0 and forces a re-run from main.c

  /**************** report some statistics only in 0-D ******************/
  if (DIM == 0 && STATS) {
	tres = endthist - (*begthist);
	Emean = meanvector(Ehist,ihstart,ih);
	Vmean = meanvector(Vhist,ihstart,ih);
	Estd = stdvector(Ehist,ihstart,ih);
	Vstd = stdvector(Vhist,ihstart,ih);
	printf("  Eavg = %.4f     +/- %.2f%% [eV]    |\n",Emean,Estd/Emean*100.);
	printf("  Vavg =% .4e +/- %.2f%% [cm/s]  |  Mobility = %.0f [cm^2/V/s]\n",
		   Vmean, fabs(Vstd/Vmean*100.), fabs(Vmean/EFIELD));
	powerPH = realsimscale*(EphLO+EphTO+EphLA+EphTA)/tres;
	printf("  Power Pho = %.3e (eV/s/cm^3)\n", fabs(powerPH));
	printf("  OpPower = %.3e [eV/s]          |", (EphLO+EphTO)/tres);
	printf("  AcPower = %.3e [eV/s]\n", (EphLA+EphTA)/tres);

	for (i = 1; i <= nelec; i++) {
	  if (iv[i] <= 2) icold += 1.;
	  else ihot += 1.;
	}
	printf("  Valley ratio = %f\n", 2.*icold/ihot);
	if (MOBCALC) MOBILITY = fabs(Vmean/EFIELD);

	fprintf(fp, "%%\n%% EFIELD     <E>      |<vx>|      PowJE      PowPH      LAPow      TAPow      LOPow      TOPow [eV/s/cm^3]\n");
	if (EFSTEPS == 0) fprintf(fp, "%%");
	// save full stat summary for each E-field, incl phonon power diss.
	fprintf(fp, " % 6.1f % .5f % .4e % .3e % .3e % .3e % .3e % .3e % .3e\n%%\n",
			EFIELD,Emean,fabs(Vmean),fabs((DOPING+Ni)*Vmean*EFIELD),
			powerPH, realsimscale*EphLA/tres, realsimscale*EphTA/tres,
			realsimscale*EphLO/tres, realsimscale*EphTO/tres);
  }

  /***** initial particle starting times if mcloop may be called again *****/
  if ((DIM == 0) && ((EFSTEPS > 0) || (STATS == 0))) {
	for (i = 1; i <= nelec; i++) Elec[TS][i] = -log(ranmc())/MAXSCAT;
	// also save useful average values for next run
	Ehist[1] = Ehist[ih];  Vhist[1] = Vhist[ih];
	// Vtimeavg = Vhist[ih];
  }

#else

  STATS = 1;    // set STATS = 1 so it won't repeat after return to main()

#endif

}
