#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "constants.h"
#include "acrate3d.c"

/******* return density of states for a given energy and eff.mass ******/
double DOSe(double Ed) {
  double sgamma, dosconst, dosE=0.0;
  if (Ed > 0.0) {
	sgamma = sqrt(Ed*(1.+alpha*Ed));
	// includes 2x for spin, else it would be 4x in the denominator
	dosconst = 2.*pi*pi*hbar*hbar*hbar;
	dosE = pow(2.*mdos, 1.5)*sgamma*(1. + alpha2*Ed)/dosconst;
  }
  return dosE;
}

void ratecalc(double Ed) {
  int i;
  double scatOconst, Nq;
  double rateLAem=0.0, rateLAabs=0.0, rateTAem=0.0, rateTAabs=0.0;
  double rateLOgem, rateLOgabs, rateTOfem, rateTOfabs;
  double rateTAgem, rateTAgabs, rateLAfem, rateLAfabs;
  double rateTAfem, rateTAfabs, rateLAgem, rateLAgabs;
  double rateDOP=0.0, ratesum, gamma, ebeta, beta, ks, a0, v0, xx, Gx, bb;
  double tempTAf=T, tempOP=T, tempAC=T;

  /************************************************************************/
  /************ INTRAVALLEY SCATTERING (ACOUSTIC LA+TA ONLY) **************/
  /************************************************************************/

  /*********** INTRAvalley (LA) phonon emission (inelastic) *************/
  if (Ed >= EminLA) rateLAem = acrate3(Ed,LA,EMS);  /* +1 emi */
  else rateLAem = 0.;

  /*********** INTRAvalley (LA) phonon absorption (inelastic) ***********/
  rateLAabs = acrate3(Ed,LA,ABS);                             /* -1 abs */

#ifdef POP
  /*********** INTRAvalley (TA) phonon emission (inelastic) *************/
  if (Ed >= EminTA) rateTAem = acrate3(Ed,TA,EMS);  /* +1 emi */
  else rateTAem = 0.;

  /*********** INTRAvalley (TA) phonon absorption (inelastic) ***********/
  rateTAabs = acrate3(Ed,TA,ABS);                             /* -1 abs */
#endif

  /************************************************************************/
  /**************** INTERVALLEY SCATTERING (LO+TO+LA+TA) ******************/
  /************************************************************************/
  // the value of the overlap integral for inter-valley scattering may
  // be included in the coupling constants (Jacoboni 1983, p. 672)

  scatOconst = pi/(2.*rho/hbar);
  /************* optical (63 meV LO-g) phonon emission ****************/
  Nq = 1./(exp(ELOg/(kb*tempOP))-1.);
  rateLOgem = Zg*DLOg*DLOg*(Nq+1.)*DOSe(Ed-ELOg)*scatOconst/ELOg;
  /************ optical (63 meV LO-g) phonon absorption ***************/
  rateLOgabs = Zg*DLOg*DLOg*Nq*DOSe(Ed+ELOg)*scatOconst/ELOg;

  Nq = 1./(exp(ETOf/(kb*tempOP))-1.);
  /************* optical (59 meV TO-f) phonon emission ****************/
  rateTOfem = Zf*DTOf*DTOf*(Nq+1.)*DOSe(Ed-ETOf+DESIGEX)*scatOconst/ETOf;
  /************ optical (59 meV TO-f) phonon absorption ***************/
  rateTOfabs = Zf*DTOf*DTOf*Nq*DOSe(Ed+ETOf+DESIGEX)*scatOconst/ETOf;

  /************* acoustic (12 meV TA-g) phonon emission ****************/
  Nq = 1./(exp(ETAg/(kb*T))-1.);
  rateTAgem = Zg*DTAg*DTAg*(Nq+1.)*DOSe(Ed-ETAg)*scatOconst/ETAg;
  /************ acoustic (12 meV TA-g) phonon absorption ***************/
  rateTAgabs = Zg*DTAg*DTAg*Nq*DOSe(Ed+ETAg)*scatOconst/ETAg;

  /************* acoustic (47 meV LA-f) phonon emission ****************/
  Nq = 1./(exp(ELAf/(kb*T))-1.);
  rateLAfem = Zf*DLAf*DLAf*(Nq+1.)*DOSe(Ed-ELAf+DESIGEX)*scatOconst/ELAf;
  /************ acoustic (47 meV LA-f) phonon absorption ***************/
  rateLAfabs = Zf*DLAf*DLAf*Nq*DOSe(Ed+ELAf+DESIGEX)*scatOconst/ELAf;

  /************* acoustic (19 meV TA-f) phonon emission ****************/
  Nq = 1./(exp(ETAf/(kb*tempTAf))-1.);
  rateTAfem = Zf*DTAf*DTAf*(Nq+1.)*DOSe(Ed-ETAf+DESIGEX)*scatOconst/ETAf;
  /************ acoustic (19 meV TA-f) phonon absorption ***************/
  rateTAfabs = Zf*DTAf*DTAf*Nq*DOSe(Ed+ETAf+DESIGEX)*scatOconst/ETAf;
  
  /************* acoustic (18 meV LA-g) phonon emission ****************/
  Nq = 1./(exp(ELAg/(kb*T))-1.);
  rateLAgem = Zg*DLAg*DLAg*(Nq+1.)*DOSe(Ed-ELAg)*scatOconst/ELAg;
  /************ acoustic (18 meV LA-g) phonon absorption ***************/
  rateLAgabs = Zg*DLAg*DLAg*Nq*DOSe(Ed+ELAg)*scatOconst/ELAg;

  /**************** Brooks-Herring impurity scattering *****************/
  // inverse of Debye screening length, some models don't include 4*pi
  if (DOPING >= 1.0e12) {
	if (Ed < DE) Ed = DE;
	beta = sqrt(DOPING/(epssi*kb*T));
	gamma = Ed*(1.+alpha*Ed);
	//ebeta = (hbar*beta)*(hbar*beta)/(2.*mdos);
	/*rateDOP = pow(2.0,2.5)*pi*DOPING/(epssi*epssi*sqrt(ebeta*mdos))*
	  sqrt(gamma)*(1.+alpha2*Ed)/(1.+4.*gamma/ebeta); */
	/*rateDOP = 2.*pow(2./mdos,1.5)*pow(pi,3.)*pow(hbar,4.)*DOPING*beta
	 *DOSe(Ed)/(epssi*epssi*((hbar*hbar*beta*beta)+8.*(mdos*gamma)));*/
	
	ks = sqrt(2.*mdos*gamma)/hbar;
	/*rateDOP = 2.*pi*DOPING*DOSe(Ed)/
	  (hbar*epssi*epssi*beta*beta*(4*ks*ks + beta*beta));*/
	/******************** Ridley impurity scattering *********************/
	//a0 = pow(pi2*DOPING,-1./3.);   // avg dist b/w ions, maybe no pi2?
	//v0 = sqrt(2.*gamma/mdos)/(1.+alpha2*Ed);
	v0 = hbar*ks/(mdos*(1.+alpha2*Ed));
	//rateDOP = v0/a0*(1.-exp(-a0*rateDOP/v0));
	
	/******************* Kosina impurity scattering **********************/
	xx = hbar*hbar*(2.*ks)*(2.*ks)/(8.*mdos*kb*T);
	// prefactors are from a numerical fit to Kosina's screening function
	Gx = (1.+0.165651*xx*xx+0.0557896*pow(xx,4.))/
	  (1.+0.829201*xx*xx+0.343064*pow(xx,4.)+0.111354*pow(xx,6.));
	bb = 4.*ks*ks/(beta*beta*Gx);
	rateDOP = DOPING/(pi2*hbar*hbar*epssi*epssi*v0*4.*ks*ks);
	rateDOP = 0.8*2.0*rateDOP*(log(1.+bb) - bb/(1.+bb));
	//        ^^^ FUDGE FACTOR TO MATCH EXPERIMENTAL MOBILITY!!!

	// FLAT rateDOP which is LARGER than all likely 1e15 - 1e20 rates
	// rateDOP = 2.2e14;  // only try this if you've got 1e20 doping in there
	// then do a LOT of rejections in scatter() where iscat == 18

	// [this is the minDOPING experiment ...]
	/*	beta = sqrt(minDOPING/(epssi*kb*T));
		bb = 4.*ks*ks/(beta*beta*Gx);
		rateminDOP = minDOPING/(pi2*hbar*hbar*epssi*epssi*v0*4.*ks*ks);
		rateminDOP = 0.8*2.0*rateDOP*(log(1.+bb) - bb/(1.+bb));
		
		rateDOP = rateDOP + rateminDOP;  // NOT right
	*/
  }

  rate[2] = rateLAem;
  rate[3] = rateLAabs;
  rate[4] = rateTAem;
  rate[5] = rateTAabs;
  rate[6] = rateLOgem;
  rate[7] = rateLOgabs;
  rate[8] = rateTOfem;
  rate[9] = rateTOfabs;
  rate[10] = rateTAgem;
  rate[11] = rateTAgabs;
  rate[12] = rateLAfem;
  rate[13] = rateLAfabs;
  rate[14] = rateTAfem;
  rate[15] = rateTAfabs;
  rate[16] = rateLAgem;
  rate[17] = rateLAgabs;
  rate[18] = rateDOP;

  ratesum = 0.;
  // return the sum of all scattering rates in rate[1] to be
  // used when figuring out MAXSCAT inside calcscatrate()
  for (i = 2; i <= NPROC; i++) ratesum += rate[i];
  rate[1] = ratesum;

  return;
}

/*********************************************************************/
/* pre-calculate scattering rates in linearly spaced matrix ScatMat
   from EMIN..EMAX with small energy intervals (.5 meV), size doesn't
   matter, i'll be using integer division (Laux) to find the energy,
   NOT expensive binary searches with locate */
void calcscatrate(void) {
  int i=1, j=1, ims=1;
  char line[MAXLINE];
  FILE *in;

  in = fopen("scatrates.txt","r");

  if ((in == NULL) || (READRATES == 0)) {
	fprintf(stderr, "Computing Scattering Rates ");

	/************** scattering processes matrix defined ****************
	 ScatMat[1..NEBINMAX][1..NPROC] but only ever used up to NEBIN     */
	for (j = 1; j <= NEBIN; j++) {
	  if ((j % 100) == 0) fprintf(stderr,".");
	  ratecalc(ebins[j]);
	  /*edist[j] = DOSe(ebins[j]);   // temporary for DOS printout */
	  for (i = 1; i <= NPROC; i++) {
		ScatMat[j][i] = rate[i];
	  }
	  // establish MAXSCAT as the max of the sum of rates (col 1)
	  if (ScatMat[j][1] > MAXSCAT) MAXSCAT = ScatMat[j][1];
	}

	for (j = 1; j <= NEBIN; j++) {       // establish self scattering
	  ScatMat[j][1] = MAXSCAT - ScatMat[j][1];
	}

	/*print2vectors(ebins, edist, 1, NEBIN, "dosa.txt");
	  exit(0);*/

	printf("\nWriting Scattering Rates [%i x %i] to File --> scatrates.txt\n",NEBIN,NPROC);
	printmatrix(ScatMat, 1, NEBIN, 1, NPROC, ".", "scatrates.txt", 1);

  } else {
	fprintf(stderr, "Reading Scattering Rates from File <-- scatrates.txt\n");
	fprintf(stderr, "If you've changed TEMP, EMAX, ALPHA or any phonon scattering parameters\nsince the last run please delete scatrates.txt and rerun monet.\n");

	while (fgets(line, MAXLINE-1, in) != NULL) {
	  if (nocomment(line,MAXLINE)) {
		ebins[j] = atof(strtok(line,"\n\t "));
		for (i = 1; i <= NPROC; i++) {
		  ScatMat[j][i] = atof(strtok(NULL,"\n\t "));
		}
	  }
	  j++;
	}
	fclose(in);
	// try to catch if EMAX has changed since last run
	if (NEBIN != (j-1)) {
	  nrerror("*** ERROR:  EMAX has changed, please delete scatrates.txt and rerun ***\n");
	}
	NEBIN = j-1;

	// figure out MAXSCAT from any line of the file that was read in...
	MAXSCAT = 0.0;
	for (i = 1; i <= NPROC; i++) MAXSCAT += ScatMat[5][i];
	// ...such as line 5
  }

  printf("MAXSCAT = %g 1/sec (1/MAXSCAT = %g sec)\n", MAXSCAT, 1/MAXSCAT);
  if (DT > 1./MAXSCAT) printf("*** WARNING:  Time Step > 1/MAXSCAT ***\n");

  return;
}


/*************************************************************************/
/*************************************************************************/
/*************************************************************************/

int intra_scat(int j, double ks, int lt, double aed, double *qdum, double *edum) {
  double C, qM, kspmax, Epmax, q0, f0, fx, ephon, acost, E0;

  assert((j >= 1) && (j <= nelec));
  E0=Elec[EE][j];

  /* aed = +1. for emission ****** aed = -1. for absorption */
  if (aed > 0.) {
	qM = 2.*ks;                 /* for emission, q <= 2*ks always */
  } else {
	Epmax = E0 + hbar*get_wq(QMAX,lt);
	kspmax = sqrt(2.*mdos*Epmax*(1.+alpha*Epmax))/hbar;
	qM = ks + kspmax;      /* for absorption, q <= this qM always */
  }

  C = qM*get_Mq(qM,lt,aed)*(1.+alpha2*(E0-aed*hbar*get_wq(qM,lt)));
  // DO ENTERS AN INFINITE LOOP HERE WHEN E IS VERY SMALL, LIKE 4E-6 so
  // << DE (5e-4) so check this small enegy against .5*mdos*vta*vta (and
  // vla) above!!! maybe store (and check against) or try get rid of it
  do {
	q0 = QMIN + (qM-QMIN)*ranmc();
	ephon = hbar*get_wq(q0,lt);
	f0 = C*ranmc();
	fx = q0*get_Mq(q0,lt,aed)*(1.+alpha2*(E0-aed*ephon));
	acost = fabs_cost(q0,ks,E0,lt,aed);
  } while ((f0 > fx) || (acost > 1.));

  /* make sure you don't get a negative phonon or emit larger than E */
  if ((E0-aed*ephon < EMIN) || (ephon < 0.)) {
	fprintf(stderr, "*** intra_scat problem! ***\n");
	intra_scat(j,ks,lt,aed,qdum,&ephon);
  }

  // set the electron energy, consistent w/the momentum exchange above
  Elec[EE][j] -= aed*ephon;

  // return the phonon energy and momentum for STATS
  *qdum = q0;
  *edum = ephon;

  return 1;
}

int inter_scat_f(int j, int lt, double aed, double *qdum, double *edum) {
  int Pxyz, cnt=0, iv0, REDO=1;
  double ephon, ephon0, qphon, Ep, bet, calp, salp;
  double pp, pxo, pyo, pzo, dpx, dpy, dpz, r, EX=0.0, E0;

  assert((j >= 1) && (j <= nelec));
  E0  = Elec[EE][j];
  iv0 = iv[j];

#ifdef POP   // store original momentum components and valley
  if (aed > 0.) assert((E0 - hbar*get_wq(QFp*QMAX,lt)) > EMIN);
  pxo = Elec[PX][j];  pyo = Elec[PY][j];  pzo = Elec[PZ][j];
#else
  if ((lt == LA) || (lt == LO)) {
	ephon = ELAf;
  } else if (lt == TA) {
	ephon = ETAf;
  } else {
	ephon = ETOf;
  }
  qphon = QFp*QMAX;
  if (aed > 0.) assert((E0 - ephon) > EMIN);
#endif

  // change valley upon scattering if we're in strained or bulk Si
  if (SIGEX > 0.0) {
#ifdef POP
	ephon = hbar*get_wq(QFp*QMAX,lt);  // MAKE SURE WHEN THIS lt IS LA OR LO
#endif
	r = ranmc();
	// DOSe(E-aed*ephon+DESIGEX);
	if ((iv0+1)/2 == 2)    {       // iv = {3,4} (D2 lower valley)
	  if (r*DOSe(E0-aed*ephon+DESIGEX) > DOSe(E0-aed*ephon-DESIGEX)) {
		return 0;                  // self-scattering and return
	  } else {
		change_noneq_valley(j);        // --> {1,2,5,6} continue
		EX = -DESIGEX;
	  }
	} else {                 // iv = {1,2,5,6} (D4 upper valley)
	  if (r < 0.5) {
		iv[j] = 3 + ifloor(2.*ranmc());        // --> {3,4}
		EX = DESIGEX;
	  } else if (r < (0.5 + DOSe(E0-aed*ephon)/DOSe(E0-aed*ephon+DESIGEX))) {
		if ((iv0+1)/2 == 1)            // iv = {1,2} --> {5,6}
		  iv[j] = 5 + ifloor(2.*r);
		else                           // iv = {5,6} --> {1,2}
		  iv[j] = 1 + ifloor(2.*r);
	  } else {
		return 0;
	  }
	}
	// DON'T FORGET TO ADD/SUBTRACT EX = +/-DESIGEX ENERGY BELOW!!!
  } else {
	change_noneq_valley(j);
  }
#ifdef POP
  dpx = pvx[iv[j]]-pxo;  dpy = pvy[iv[j]]-pyo;  dpz = pvz[iv[j]]-pzo;
  dpx = PQMAX-fabs(dpx); dpy = PQMAX-fabs(dpy); dpz = PQMAX-fabs(dpz);
  qphon = sqrt(dpx*dpx + dpy*dpy + dpz*dpz)/hbar;
#endif
  
  /* to do: for f-scat with LA/LO phonons, choose if it's LA or LO based
	 on a random number weighed by the occupation of those branches i.e.
	 Nq(LA) vs Nq(LO) for whatever qphon seems to satisfy the particular
	 transition.  maybe shoot an email to Ravaioli?
   */

  // new longitudinal axis momentum component
  Pxyz = PX-1+(int)(iv[j]+1)/2;
#ifdef POP
  if (qphon < QMIN) fprintf(stderr, "WARNING: qphon = %g < QMIN\n",qphon/QMAX);
  // original estimate on phonon energy
  ephon = hbar*get_wq(qphon,lt);
  // remember inside get_wq() if qphon > QMAX then phonon is LO not LA

  do {
#endif
	ephon0 = ephon;
	// estimate new electron energy
	Ep = E0 - aed*ephon0 + EX;
#ifdef POP
	if (Ep > EMIN) {
#endif
	  // compute new momentum in Herring-Vogt space
	  pp = sqrt(2.*mdos*Ep*(1.+alpha*Ep));

	  do {
		// choose the set of random angles
		calp = 1.-2.*ranmc();
		salp = sqrt(1.-calp*calp);
		bet = pi2*ranmc();
		// new momentum components in new valley iv[j]
		Elec[PX][j] = pvx[iv[j]] + sqrt(mx[iv[j]]/mdos)*pp*salp*cos(bet);
		Elec[PY][j] = pvy[iv[j]] + sqrt(my[iv[j]]/mdos)*pp*salp*sin(bet);
		Elec[PZ][j] = pvz[iv[j]] + sqrt(mz[iv[j]]/mdos)*pp*calp;
	  } while ((fabs(Elec[Pxyz][j]) > PQMAX) && (BZCLIP));

#ifdef POP
	  // difference between final & initial momentum vectors
	  dpx = Elec[PX][j]-pxo;  dpy = Elec[PY][j]-pyo;  dpz = Elec[PZ][j]-pzo;
	  // get qphon = norm of the resultant q-vector reduced to 1st BZ
	  dpx = PQMAX-fabs(dpx); dpy = PQMAX-fabs(dpy); dpz = PQMAX-fabs(dpz);
	  qphon = sqrt(dpx*dpx + dpy*dpy + dpz*dpz)/hbar;  // always positive
	  if (qphon < QMIN) fprintf(stderr, "WARNING: qphon = %g < QMIN\n",qphon/QMAX);
	  // get resulting phonon energy and momentum
	  ephon = hbar*get_wq(qphon,lt);
	  // re-estimate new electron energy
	  Ep = E0 - aed*ephon0 + EX;
	  // upper limit for cnt = 200, see right below
	  cnt++;
	  if ((Ep > EMIN) && ((fabs(ephon-ephon0)/ephon0) < 0.05)) REDO = 0;
	} else {
	  // go back to basic estimate on phonon energy
	  qphon = QFp*QMAX;
	  ephon = hbar*get_wq(qphon,lt);
	}

	// allow it to break out if cnt goes too high
	if ((cnt > 200) && REDO) {
	  REDO = 0;
	  qphon = QFp*QMAX;
	  ephon = hbar*get_wq(qphon,lt);
	}
	// empirical note: this happens VERY rarely so it's OK

	// reject several unwanted scenarios and tighten ephon guess
  } while (REDO);
#endif

  // set the electron energy, consistent w/the momentum exchange above
  Elec[EE][j] = E0 - aed*ephon0 + EX;

  if (aed > 0.) assert(Elec[EE][j] > EMIN);
  // return the phonon energy and momentum for STATS
  *qdum = qphon;
  *edum = ephon;
  // this should be consistent within 5% with the electron energy
  // exchange [based on ephon0] but here i'm keeping qphon + ephon
  // for consistency within the generated phonon statistics
  
  if (cnt > 200) fprintf(stderr, "cnt = %i\n", cnt);
  return 1;
}

int inter_scat_g(int j, int lt, double aed, double *qdum, double *edum) {
  int Pxyz, cnt=0, iv0, REDO=1;
  double ephon, ephon0, qphon, qphon0, Ep, bet, calp, salp;
  double pp, pxo, pyo, pzo, dpx, dpy, dpz, r, E0;

  assert ((j >= 1) && (j <= nelec));
  E0  = Elec[EE][j];
  iv0 = iv[j];

#ifdef POP   // store original momentum components
  if (aed > 0.) assert((E0 - hbar*get_wq(QGp*QMAX,lt)) > EMIN);
  pxo = Elec[PX][j];  pyo = Elec[PY][j];  pzo = Elec[PZ][j];
#else
  if (lt == LA) {
	ephon = ELAg;
  } else if (lt == TA) {
	ephon = ETAg;
  } else {
	ephon = ELOg;
  }
  if (aed > 0.) assert((E0 - ephon) > EMIN);
  qphon = QGp*QMAX;
#endif

  change_equiv_valley(j);
  // new longitudinal axis momentum component
  Pxyz = PX-1+(int)(iv[j]+1)/2;

#ifdef POP
  dpx = pvx[iv[j]]-pxo;  dpy = pvy[iv[j]]-pyo;  dpz = pvz[iv[j]]-pzo;
  qphon0 = fabs(2.*PQMAX-sqrt(dpx*dpx + dpy*dpy + dpz*dpz))/hbar;
  if (qphon0 < QMIN) fprintf(stderr, "WARNING: qphon0 = %g < QMIN\n",qphon0/QMAX);
  // original estimate on phonon energy
  ephon = hbar*get_wq(qphon0,lt);

  do {
#endif
	ephon0 = ephon;
	// estimate new electron energy
	Ep = E0 - aed*ephon0;
#ifdef POP
	if (Ep > EMIN) {
#endif
	  // compute new momentum in Herring-Vogt space
	  pp = sqrt(2.*mdos*Ep*(1.+alpha*Ep));

	  do {
		// choose the set of random angles
		calp = 1.-2.*ranmc();
		salp = sqrt(1.-calp*calp);
		bet = pi2*ranmc();
		// new momentum components in new valley iv[j]
		Elec[PX][j] = pvx[iv[j]] + sqrt(mx[iv[j]]/mdos)*pp*salp*cos(bet);
		Elec[PY][j] = pvy[iv[j]] + sqrt(my[iv[j]]/mdos)*pp*salp*sin(bet);
		Elec[PZ][j] = pvz[iv[j]] + sqrt(mz[iv[j]]/mdos)*pp*calp;
	  } while ((fabs(Elec[Pxyz][j]) > PQMAX) && (BZCLIP));

#ifdef POP
	  // difference between final & initial momentum vectors
	  dpx = Elec[PX][j]-pxo;  dpy = Elec[PY][j]-pyo;  dpz = Elec[PZ][j]-pzo;
	  // get qphon = norm of the resultant q-vector reduced to 1st BZ
	  qphon = fabs(2.*PQMAX-sqrt(dpx*dpx + dpy*dpy + dpz*dpz))/hbar;
	  if (qphon < QMIN) fprintf(stderr, "WARNING: qphon = %g < QMIN\n",qphon/QMAX);
	  // get resulting phonon energy and momentum
	  ephon = hbar*get_wq(qphon,lt);
	  // estimate new electron energy
	  Ep = E0 - aed*ephon;
	  // upper limit for cnt = 200, see right below
	  cnt++;
	  if ((Ep > EMIN) && ((fabs(ephon-ephon0)/ephon0) < 0.05)) REDO = 0;
	} else {
	  // original estimate on phonon energy
	  qphon = QGp*QMAX;
	  ephon = hbar*get_wq(qphon,lt);
	}
	// allow it to break out if cnt goes too high
	if ((cnt > 200) && REDO) {
	  REDO = 0;
	  qphon = QGp*QMAX;
	  ephon = hbar*get_wq(qphon,lt);
	}
	// empirical note: this happens VERY rarely so it's OK

  } while (REDO);
#endif

  // set the electron energy, consistent w/the momentum exchange above
  Elec[EE][j] = E0 - aed*ephon;

  if (aed > 0.) assert(Elec[EE][j] > EMIN);
  // return the phonon energy and momentum for STATS
  *qdum = qphon;
  *edum = ephon;
  // consistent within 5% with the electron energy exchange [based on
  // ephon0] but here i'm keeping qphon + ephon for consistency within
  // the generated phonon statistics
  
  //if (cnt > 200) fprintf(stderr, "cnt = %i\n", cnt);
  return 1;
}

/********** Recall scattering process index 1..NPROC *************
   1 = self scattering
   2 = LA intravalley emission
   3 = LA intravalley absorption
   4 = TA intravalley emission
   5 = TA intravalley absorption
   6 = LO-g intervalley emission
   7 = LO-g intervalley absorption
   8 = TO-f intervalley emission
   9 = TO-f intervalley absorption
  10 = TA-g intervalley emission
  11 = TA-g intervalley absorption
  12 = LA-f intervalley emission
  13 = LA-f intervalley absorption
  14 = TA-f intervalley emission
  15 = TA-f intervalley absorption
  16 = LA-g intervalley emission
  17 = LA-g intervalley absorption
  18 = ionized impuring scattering                               */
/*****************************************************************/
   
void scatter(int j) {
  int ret, Pxyz, iscat=1, iebin, ipx=1, ipy=1, cnt=0;
  double cumprob, r0, r1, pprime, salp, calp, bet, ephon, ks;
  double kprime, cbet, sbet, q0, phi, cphi, sphi, eta, ceta, seta;
  double cgam, sgam, cgamp, cgampp, padjust, wavg1, wavg2;
  double locdop, locrate, Gx, bb, xx, v0, beta;

  assert ((j >= 1) && (j <= nelec));

  /* Elec[EE][j] = energy(j);        // do this if changing drift() below */

  /***** complain or re-adjust if energy out of program range *******/
  if (Elec[EE][j] < EMIN) {
	// complain and stop if the energy is below EMIN
	fprintf(stderr, "SCAT1:  E[%i] = %g < EMIN!\n", j, Elec[EE][j]);
	fprintf(stderr, "momenta: [%g %g %g]\n",
			Elec[PX][j]/PQMAX, Elec[PY][j]/PQMAX, Elec[PZ][j]/PQMAX);
	exit(-1);
  } else if (Elec[EE][j] > EMAX) {
	// readjust if energy above EMAX, may need to track energy "lost" here
	iebin = NEBIN;
	padjust = sqrt((EMAX*(1.+alpha*EMAX))/(Elec[EE][j]*(1.+alpha*Elec[EE][j])));
	Elec[PX][j] = pvx[iv[j]] + (Elec[PX][j]-pvx[iv[j]])*padjust;
	Elec[PY][j] = pvy[iv[j]] + (Elec[PY][j]-pvy[iv[j]])*padjust;
	Elec[PZ][j] = pvz[iv[j]] + (Elec[PZ][j]-pvz[iv[j]])*padjust;
	Elost += (Elec[EE][j] - EMAX);   // add up "lost" energy
	Elec[EE][j] = EMAX;
	wavg1 = 1.;
	wavg2 = 0.;
  } else {
	// integer division to locate the LEFT (low) end of the energy bin
	iebin = 1 + ifloor(Elec[EE][j]/DE);
	wavg1 = (Elec[EE][j] - ebins[iebin])/DE;
	wavg2 = 1. - wavg1;
  }

  if (iebin > NEBIN) nrerror("ERROR:  Out of Energy Range!\n");

  /************* convert to spherical Herring-Vogt space ************/
  ks = sqrt(2.*mdos*Elec[EE][j]*(1.+alpha*Elec[EE][j]))/hbar;
  if (2.*ks > QMAX) nrerror("ERROR:  Large electron K-vector!\n");

  /********** identify scattering process (integer iscat) ***********/
  /*  if (FASTSCAT) 
	r1 = ranmc()*MScat[iebin];
	else */
  r0 = ranmc();
  r1 = r0*MAXSCAT;
  // start with iscat = 1 which is for self-scattering
  cumprob = wavg1*ScatMat[iebin][iscat] + wavg2*ScatMat[iebin+1][iscat];
  while ((r1 > cumprob) && (iscat < NPROC)) {
	iscat++;
	// catch any emission errors that may be introduced by scat-rate gridding
	if ((iscat == 2) && (Elec[EE][j] < EminLA)) {         // no good LA emission
	  cumprob += 0.0;
	} else if ((iscat == 4) && (Elec[EE][j] < EminTA)) {  // no good TA emission
	  cumprob += 0.0;
	} else if ((iscat == 6) && (Elec[EE][j] < ELOg)) {    // no good LOg emission
	  cumprob += 0.0;
	} else if ((iscat == 8) && (Elec[EE][j] < ETOf)) {    // no good TOf emission
	  cumprob += 0.0;
	} else if ((iscat == 10) && (Elec[EE][j] < ETAg)) {   // no good TAg emission
	  cumprob += 0.0;
	} else if ((iscat == 12) && (Elec[EE][j] < ELAf)) {   // no good LAf emission
	  cumprob += 0.0;
	} else if ((iscat == 14) && (Elec[EE][j] < ETAf)) {   // no good TAf emission
	  cumprob += 0.0;
	} else if ((iscat == 16) && (Elec[EE][j] < ELAg)) {   // no good LAg emission
	  cumprob += 0.0;
	} else {
	  cumprob += (wavg1*ScatMat[iebin][iscat] + wavg2*ScatMat[iebin+1][iscat]);
	}
  }

  assert((iscat >= 1) && (iscat <= NPROC));

  if (iscat == 1) {            /*** nothing, just self-scattering ***/
	isf++;
	return;
  } else {
	/******* find out where (on grid) the scattering took place *****/
	if (DIM) {
	  ipx = Gxp[j];
	  if (DIM == 2) ipy = Gyp[j];
	  /******** pre-run cloud-in-cell to get weights now ********/
	  cloud_in_cell(j, ipx, ipy);
	}

	/**************** INTRA-valley acoustic phonons ******************/
	if (iscat == 2) {             /***** LA emission (inelastic) *****/
	  intra_scat(j, ks, LA, EMS, &q0, &ephon);
#ifndef NOISE   	  // collect data after transient
	  if (STATS) phonon_stats(EMS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
	  iae++;
#endif

	} else if (iscat == 3) {      /***** LA absorption (inelastic) ***/
	  intra_scat(j, ks, LA, ABS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
	  iaa++;
#endif

	} else if (iscat == 4) {      /***** TA emission (inelastic) *****/
	  intra_scat(j, ks, TA, EMS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
	  iae++;
#endif

	} else if (iscat == 5) {      /***** TA absorption (inelastic) ***/
	  intra_scat(j, ks, TA, ABS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
	  iaa++;
#endif

	  /********************************************************************/
	  /*********** INTER-valley acoustic AND optical phonons **************/
	  /********************************************************************/
	} else if (iscat == 6) {    /* LO-g emission */
	  ret = inter_scat_g(j, LO, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphLO,NELph0,NOgrid,Ogrid,ipx,ipy);
	  ioe++;
#endif

	} else if (iscat == 7) {    /* LO-g absorption */
	  ret = inter_scat_g(j, LO, ABS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphLO,NELph0,NOgrid,Ogrid,ipx,ipy);
	  ioa++;
#endif

	} else if (iscat == 8) {    /* TO-f emission */
	  ret = inter_scat_f(j, TO, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphTO,NETph0,NOgrid,Ogrid,ipx,ipy);
	  ioe++;
#endif

	} else if (iscat == 9) {    /* TO-f absorption */
	  ret = inter_scat_f(j, TO, ABS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphTO,NETph0,NOgrid,Ogrid,ipx,ipy);
	  ioa++;
#endif

	  //
	  // DOES ret == 0 MEAN IT CANNOT EMIT PHONONS OF SHORTER Q (IN
	  // THE REDUCED BZ) AND THEREFORE OF LOWER ENERGY?!  DOESN'T MAKE
	  // MUCH SENSE... NAH.  BUT CHECK ON RET ANYWAY.
	  //

	} else if (iscat == 10) {    /* TA-g emission */
	  ret = inter_scat_g(j, TA, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 11) {    /* TA-g absorption */
	  ret = inter_scat_g(j, TA, ABS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 12) {    /* LA-f emission */
	  ret = inter_scat_f(j, LA, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) {
		if (ephon < (hbar*get_wq(QMAX,LA))) {   // can also use q0 < QMAX
		  phonon_stats(EMS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
		} else {
		  phonon_stats(EMS,ephon,&EphLO,NELph0,NOgrid,Ogrid,ipx,ipy);
		}
	  }
#endif

	} else if (iscat == 13) {    /* LA-f absorption */
	  ret = inter_scat_f(j, LA, ABS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) {
		if (ephon < (hbar*get_wq(QMAX,LA))) {   // can also use q0 < QMAX
		  phonon_stats(ABS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
		} else {
		  phonon_stats(ABS,ephon,&EphLO,NELph0,NOgrid,Ogrid,ipx,ipy);
		}
	  }
#endif

	} else if (iscat == 14) {    /* TA-f emission */
	  ret = inter_scat_f(j, TA, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 15) {    /* TA-f absorption */
	  ret = inter_scat_f(j, TA, ABS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphTA,NETph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 16) {    /* LA-g emission */
	  ret = inter_scat_g(j, LA, EMS, &q0, &ephon);
	  if (ret == 0) { isf++; return; }  // treat as self-scattering
#ifndef NOISE
	  if (STATS) phonon_stats(EMS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 17) {    /* LA-g absorption */
	  ret = inter_scat_g(j, LA, ABS, &q0, &ephon);
#ifndef NOISE
	  if (STATS) phonon_stats(ABS,ephon,&EphLA,NELph0,NAgrid,Agrid,ipx,ipy);
#endif

	} else if (iscat == 18) {    /* ionized impurity scattering */

	  if (DIM) {    	  // interpolate local doping concentration
		locdop = wcic[1]*DOPgrid[ipx] + wcic[2]*DOPgrid[ipx+1];
		if (DIM == 2) {
		  locdop += wcic[3]*DOPgrid[ipx+nyg] + wcic[4]*DOPgrid[ipx+nyg+1];
		}

		if (locdop < 0.999999*DOPING) {
		  // Kosina impurity scattering with locdop instead of DOPING
		  // [the 0.999999 is there to avoid some numerical round-off errors]
		  beta = sqrt(locdop/(epssi*kb*T));
		  v0 = hbar*ks/(mdos*(1.+alpha2*Elec[EE][j]));
		  xx = hbar*hbar*(2.*ks)*(2.*ks)/(8.*mdos*kb*T);
		  Gx = (1.+0.165651*xx*xx+0.0557896*pow(xx,4.))/
			(1.+0.829201*xx*xx+0.343064*pow(xx,4.)+0.111354*pow(xx,6.));
		  bb = 4.*ks*ks/(beta*beta*Gx);
		  locrate = locdop/(pi2*hbar*hbar*epssi*epssi*v0*4.*ks*ks);
		  locrate = 0.8*2.0*locrate*(log(1.+bb) - bb/(1.+bb));

		  /*
			fprintf(stderr, "\nx = %g .. locdop = %g .. DOPING = %g\n",
			Elec[X][j]*1e7, locdop, DOPING);
			fprintf(stderr, "E = %g .. locrate = %g .. Scatrate = %g\n", Elec[EE][j],
			locrate, wavg1*ScatMat[iebin][iscat] + wavg2*ScatMat[iebin+1][iscat]);
		  */
		
		  /** reject some cases when the local doping (and scattering rate)
			  is less than the maxim doping and scat rate, respectively **/
		  r0 = ranmc();
		  if (r0 > (locrate/(wavg1*ScatMat[iebin][iscat] +
							 wavg2*ScatMat[iebin+1][iscat]))) {
			return;
		  }
		}
	  }
	  
	  /********* compute ISOTROPIC new angles for Kosina's model **********/
	  pprime = sqrt(2.*mdos*Elec[EE][j]*(1.+alpha*Elec[EE][j]));
	  Pxyz = PX-1+(int)(iv[j]+1)/2;
	  do {
		calp = 1.-2.*ranmc();
		salp = sqrt(1.-calp*calp);
		bet = pi2*ranmc();
		/********** compute new isotropic momentum components *************/
		Elec[PX][j] = pvx[iv[j]] + sqrt(mx[iv[j]]/mdos)*pprime*salp*cos(bet);
		Elec[PY][j] = pvy[iv[j]] + sqrt(my[iv[j]]/mdos)*pprime*salp*sin(bet);
		Elec[PZ][j] = pvz[iv[j]] + sqrt(mz[iv[j]]/mdos)*pprime*calp;
	  } while ((fabs(Elec[Pxyz][j]) > PQMAX) && (BZCLIP));
	  return;

	} else {
	  fprintf(stderr,"*** ERROR:  Illegal Scattering Mechanism ***\n");
	  exit(-1);
	}

	/******** energy conservation for magnitude of new momentum *******/
	if (Elec[EE][j] < EMIN) {
	  fprintf(stderr, "SCAT2:  E = %g < EMIN!\n", Elec[EE][j]);
	  fprintf(stderr, "SCAT2:  iscat = %i\n", iscat);
	  exit(-1);
	}

	/********* non-parabolicity from tomizawa eq. 1.11 page 8 *********/
	pprime = sqrt(2.*mdos*Elec[EE][j]*(1.+alpha*Elec[EE][j]));
	// this is the longitudinal axis momentum component
	Pxyz = PX-1+(int)(iv[j]+1)/2;

	/******** intra-valley acoustic, NON-ISOTROPIC final angle ********/
	if (iscat <= 5) {
	  kprime = pprime/hbar;
	  /***** momentum conservation for angle beta b/w ks and kprime *****/
	  cbet = (ks*ks+kprime*kprime-q0*q0)/(2.*ks*kprime);
	  /****************** detect if angle is foobar-ed ******************/
	  if (fabs(cbet) > 1.) {
		fprintf(stderr,"SCAT3: E = %g, iscat = %i\n", Elec[EE][j], iscat);
		fprintf(stderr,"SCAT3: ks = %g, kprime = %g, q0 = %g\n",ks,kprime,q0);
		fprintf(stderr,"SCAT3: cbet = %g  cos(ks,q) = %g\n",
				cbet, fabs_cost(q0,ks,Elec[EE][j],1,-1.));
		fprintf(stderr,"SCAT3: ephon = %g\n", ephon);
		exit(-1);
	  }
	  sbet = sqrt(1.-cbet*cbet);

	  /*** gam = angle b/w ks and Z-axis ***/
	  cgam = (Elec[PZ][j]-pvz[iv[j]])*sqrt(mdos/mz[iv[j]])/(hbar*ks);
	  if (fabs(cgam) > 1.) {
		fprintf(stderr, "\n*** ERROR: cos(gamma) > 1 (angle b/w ks and z-axis) ***\n");
		if (DIM) fprintf(stderr, "x=%g, y=%g (nm)\n", Elec[X][j]*1e7, Elec[Y][j]*1e7);
		fprintf(stderr, "MOMT[%i]: [%g %g %g] iscat=%i, iv=%i\n",j,
				Elec[PX][j]/PQMAX, Elec[PY][j]/PQMAX, Elec[PZ][j]/PQMAX,
				iscat, iv[j]);
		fprintf(stderr, "ephon=%g, qphon=%g\n",ephon,q0/QMAX);
		fprintf(stderr, "E-stored=%g, E-calculated=%g\n", Elec[EE][j], energy(j));
		fprintf(stderr, "ks=%g, kprime=%g\n",ks/QMAX,kprime/QMAX);
		write_output_distro("initial.txt");
		exit(-1);
	  }
	  sgam = sqrt(1.-cgam*cgam);
	  /*** gamp = angle b/w ks and Y-axis ***/
	  cgamp = (Elec[PY][j]-pvy[iv[j]])*sqrt(mdos/my[iv[j]])/(hbar*ks);
	  /*** gampp = angle b/w ks and X-axis ***/
	  cgampp = (Elec[PX][j]-pvx[iv[j]])*sqrt(mdos/mx[iv[j]])/(hbar*ks);
	  
	  /*** angle b/w ks projection on X-Y plane and Y-axis ***/
	  eta = atan((Elec[PX][j]-pvx[iv[j]])*my[iv[j]]/
				 ((Elec[PY][j]-pvy[iv[j]])*mx[iv[j]]));
	  ceta = cos(eta);
	  seta = sin(eta);
	  
	  do {
		/*** rotation phi chosen at random ***/
		phi = pi2*ranmc();
		sphi = sin(phi);
		cphi = cos(phi);
		/*** compute final state momentum components, like Jacoboni Rev. Mod.
			 1983, eq. 3.62-3.64, but he has X and Z components mixed up ***/
		Elec[PZ][j] = (cbet*cgam   +sbet*cphi*sgam)*
		  sqrt(mz[iv[j]]/mdos)*pprime + pvz[iv[j]];
		Elec[PY][j] = (cbet*cgamp  -sbet*cphi*cgam*ceta -sbet*sphi*seta)*
		  sqrt(my[iv[j]]/mdos)*pprime + pvy[iv[j]];
		Elec[PX][j] = (cbet*cgampp -sbet*cphi*cgam*seta +sbet*sphi*ceta)*
		  sqrt(mx[iv[j]]/mdos)*pprime + pvx[iv[j]];
		/*** increment counter ***/
		cnt++;
	  } while ((fabs(Elec[Pxyz][j]) > PQMAX) && (cnt < 200) && (BZCLIP));
	  // if BZCLIP = 0 this && is always FALSE so it only goes thru ONCE

	  // if the counter exceeds 200 reps then make it isotropic

	  if (cnt >= 200) {          // note:  this never happens if BZCLIP = 0
		/***** compute ISOTROPIC new angles if all else fails above ******/
		do {
		  calp = 1.-2.*ranmc();
		  salp = sqrt(1.-calp*calp);
		  bet = pi2*ranmc();
		  /********** compute new isotropic momentum components *************/
		  Elec[PX][j] = pvx[iv[j]] + sqrt(mx[iv[j]]/mdos)*pprime*salp*cos(bet);
		  Elec[PY][j] = pvy[iv[j]] + sqrt(my[iv[j]]/mdos)*pprime*salp*sin(bet);
		  Elec[PZ][j] = pvz[iv[j]] + sqrt(mz[iv[j]]/mdos)*pprime*calp;
		} while ((fabs(Elec[Pxyz][j]) > PQMAX) && (BZCLIP));
	  }
	}
	
	// no need to get velocity here, do it before/after vxold in drift*()
	// if (FASTSCAT) GE[j] = 1 + ifloor(Elec[EE][j]/DE);
	return;
  }
}
